from typing import Dict, Any, NamedTuple, Union, List, TYPE_CHECKING

from core.model.factory import build_factory
from core.model.registered import Registered

if TYPE_CHECKING:
    from scenarios.scenario_models.history import Event

formatters = Registered()
formatters_factory = build_factory(formatters)


class EventFormatter:

    def __init__(self, *args, **kwargs):
        pass

    def format(self, events: List['Event']) -> List[Union[NamedTuple, Dict[str, Any]]]:
        raise NotImplementedError


class HistoryEventFormatter(EventFormatter):

    def format(self, events: List['Event']) -> List[Dict[str, Any]]:
        result = []
        for no, event in enumerate(events):
            formatted_event = self._format_event(event)
            formatted_event["no"] = no + 1
            result.append(formatted_event)
        return result

    def _format_event(self, event: 'Event') -> Dict[str, Any]:
        return {
            "scenarioName": event.scenario,
            "scenarioVersion": event.scenario_version,
            "result": event.result,
            "eventType": event.type,
            "eventContent": event.content
        }



import unittest

from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems


class MockFactory:
    def __init__(self, id, items):
        self.id = id
        self.data = items["data"]
        self.version = items["version"]


class SmartUpdatableLazyDescriptionsTest(unittest.TestCase):
    def setUp(self):
        self.descr = SmartUpdatableDescriptionsItems(MockFactory, dict(id1={"data": "raw_data_value1", "version": 0},
                                                                       id2={"data": "raw_data_value2", "version": 0}))

    def test_update_data1(self):
        update_item = dict(id1={"data": "raw_data_value3", "version": 1})
        self.descr.update_data(update_item)
        expected = "raw_data_value3"
        obj1 = self.descr["id1"]
        self.assertEqual(obj1.data, expected)



from functools import cached_property
from typing import Union

from core.model.factory import factory
from scenarios.scenario_models.history import EventFormatter


class HistoryDescription:
    enabled: bool
    event_expiration_delay: Union[int, float]

    EVENT_EXPIRATION_DELAY = 60

    def __init__(self, items, id=None):
        self.id = id
        self.enabled = items.get("enabled", False)
        self.event_expiration_delay = items.get("event_expiration_delay", self.EVENT_EXPIRATION_DELAY)
        self._formatter = items.get("formatter")

    @cached_property
    @factory(EventFormatter)
    def formatter(self) -> EventFormatter:
        return self._formatter



from typing import List, Type


class PostprocessMainLoop:

    def postprocess(self, user, message, *args, **kwargs):
        pass


class PostprocessCompose(PostprocessMainLoop):
    postprocessors: List[PostprocessMainLoop] = []

    def postprocess(self, user, message, *args, **kwargs):
        for processor in self.postprocessors:
            processor.postprocess(user, message, *args, **kwargs)


def postprocessor_compose(*args: List[Type[PostprocessMainLoop]]):
    class Compose(PostprocessCompose):
        postprocessors = [processor_cls() for processor_cls in args]
    return Compose



import importlib

from core.logging.logger_utils import log


def activate_plugins(app_config, manager):
    for plugin in app_config.PLUGINS:
        try:
            plugin = importlib.import_module(plugin)
        except ImportError as exc:
            log(f"Error due importing {plugin}.\n {exc}. Plugin {plugin} is installed?", level="ERROR")
            continue

        try:
            plugin.on_startup(app_config, manager)
        except AttributeError:
            log(f"Plugin {plugin.__name__} does not have on_startup method", level="ERROR")
            continue
        except Exception as exc:
            log(f"Error due executing {plugin.__name__} plugin.\n {exc}", level="ERROR")
            continue



import json
from functools import cached_property

from core.logging.logger_utils import log
from core.model.field import Field
from core.model.base_user import BaseUser

from scenarios.scenario_models.scenario_models import ScenarioModels
from scenarios.scenario_models.forms.forms import Forms
from scenarios.user.last_fields.last_fields import LastFields
from scenarios.user.last_scenarios.last_scenarios import LastScenarios
from scenarios.user.preprocessing_messages.prepricessing_messages_for_scenarios import \
    PreprocessingScenariosMessages
from scenarios.scenario_models.history import History
from scenarios.behaviors.behaviors import Behaviors
from scenarios.user.reply_selector.reply_selector import ReplySelector
from core.monitoring.monitoring import monitoring
import scenarios.logging.logger_constants as log_const


class User(BaseUser):
    forms: Forms
    last_fields: LastFields
    last_scenarios: LastScenarios
    scenario_models: ScenarioModels
    preprocessing_messages_for_scenarios: PreprocessingScenariosMessages
    behaviors: Behaviors
    history: History

    def __init__(self, id, message, db_data, settings, descriptions, parametrizer_cls, load_error=False):
        self.settings = settings
        try:
            user_values = json.loads(db_data) if db_data else None
        except ValueError:
            user_values = None
            monitoring.counter_load_error(settings.app_name)
            log(f"%(class_name)s.__init__ User load ValueError %(uid)s with user data {db_data}",
                params={log_const.KEY_NAME: log_const.FAILED_DB_INTERACTION,
                        "uid": str(id)}, level="ERROR")
            load_error = True
        super().__init__(id, message, user_values, descriptions, load_error)
        self.__parametrizer_cls = parametrizer_cls
        self.do_not_save = False
        self.initial_db_data = db_data

    def _initialize(self):
        db_version = self.private_vars.get(self.USER_DB_VERSION, default=0)
        self.private_vars.set(self.USER_DB_VERSION, db_version + 1)
        log("%(class_name)s.__init__ USER %(uid)s LOAD db_version = %(db_version)s.", self,
            params={"db_version": str(db_version),
                    "uid": str(self.id), log_const.KEY_NAME: "user_load"})
        self.behaviors.initialize()

    @property
    def fields(self):
        return super().fields + [Field("forms", Forms, self.descriptions["forms"]),
                                 Field("last_fields", LastFields),
                                 Field("last_scenarios", LastScenarios, self.descriptions["last_scenarios"]),
                                 Field("scenario_models", ScenarioModels, self.descriptions["scenarios"]),
                                 Field("preprocessing_messages_for_scenarios", PreprocessingScenariosMessages,
                                       self.descriptions["preprocessing_messages_for_scenarios"]),
                                 Field("behaviors", Behaviors, self.descriptions["behaviors"]),
                                 Field("history", History, self.descriptions["history"]),
                                 Field("gender_selector", ReplySelector)]

    @cached_property
    def parametrizer(self):
        return self.__parametrizer_cls(self, {})

    def expire(self):
        super().expire()
        self.behaviors.expire()
        self.forms.expire()
        self.last_fields.expire()
        self.last_scenarios.expire()
        self.history.expire()



# coding: utf-8

ADJ_FORM = "adj_form"
ADJECTIVAL_NUMBER = "adjectival_number"
AMBIGUOUS = "ambiguous"
ANIMACY = "animacy"
ANTECEDENT_LEMMA = "antecedent_lemma"
ASPECT = "aspect"
CAPITAL = "capital"
CASE = "case"
COMPOSITE_TOKEN_LENGTH = 'composite_token_length'
COMPOSITE_TOKEN_TYPE = "composite_token_type"
COMPOSITE_TOKEN_VALUE = "composite_token_value"
DEGREE = "degree"
DEPENDENCY_TYPE = "dependency_type"
DEPENDENT_ADPOSITIONS = "dependent_adpositions"
FULL_NAMES = "full_names"
GENDER = "gender"
GRAMMEM_INFO = "grammem_info"
HEAD = "head"
IMPORTANT_SURNAME = "important_surname"
IMPORTANT_SURNAME_VALUE = "important_surname_value"
IS_BASTARD = "is_bastard"
IS_BEGINNING_OF_ANAPHOR = "is_beginning_of_anaphor"
IS_BEGINNING_OF_COMPOSITE = "is_beginning_of_composite"
IS_STOP_WORD = "is_stop_word"
LATITUDE = "latitude"
LEMMA = "lemma"
LIST_OF_DEPENDENTS = "list_of_dependents"
LIST_OF_TOKEN_TYPES_DATA = "list_of_token_types_data"
LOCALITY_TYPE = "locality_type"
LOCATIVE_VALUE = "locative_value"
LONGITUDE = "longitude"
MOOD = "mood"
NUMBER = "number"
NUMFORM = "numform"
OTHER = "other"
PART_OF_SPEECH = "part_of_speech"
PERSON = "person"
RAW_GRAM_INFO = "raw_gram_info"
SAO_TYPE = "sao_type"
TENSE = "tense"
TEXT = "text"
RAW_TEXT = 'raw_text'
TIMEZONE = "timezone"
CURRENCY = 'currency'
COUNTRY = 'country'
COUNTRY_HIDDEN = 'country_hidden'
TOKEN_TYPE = "token_type"
TOKEN_VALUE = "token_value"
TRANSITIVITY = "transitivity"
VALUE = "value"
VERBFORM = "verbform"
VOICE = "voice"
ANAPHORA_ORIGINAL_LEMMA = "anaphora_original_lemma"
W2V_STRING = "w2v_string"
ORG_TYPES = "org_types"
ORGANIZATIONS = "organizations"
START_GEO_SEQUENCE = "start_geo_sequence"
START_ORG_SEQUENCE = "start_org_sequence"
SEQUENCE_TYPE = "sequence_type"
SEQUENCE_ITEM_VALUE = "sequence_item_value"
PART_OF_SPEECH_PYMORPHY = "_POS"



ANNOTATIONS = "annotations"
MESSAGE = "message"
APPLICATION_ID = "applicationId"
APP_INFO = "app_info"
APP_VERSION_ID = "appversionId"
AUTO_LISTENING = "auto_listening"
CLASSES = 'classes'
DEVICE = "device"
FINISHED = "finished"
FRONTEND_ENDPOINT = "frontendEndpoint"
FRONTEND_TYPE = "frontendType"
INTENT = "intent"
INTENT_META = "intent_meta"
ITEMS = "items"
MESSAGE_NAME = "messageName"
PROBAS = 'probas'
PROJECT_ID = "projectId"
PROJECT_NAME = "projectName"
PRONOUNCE_TEXT = "pronounceText"
SERVER_ACTION = "server_action"
SMART_BIO = "smartBio"
SUB = "sub"
SYSTEM_NAME = "systemName"
USER_CHANNEL = "userChannel"
USER_ID = "userId"
MESSAGE = "message"

DEBUG_INFO = "debug_info"
DEBUG_INFO_APP_KEY = "debug_info_key"
EVENTS = "events"
HISTORY_DATA = "historyData"



import logging
from unittest import TestCase
from unittest.mock import Mock

from core.logging.logger_utils import log


class TestLogger(TestCase):

    def test_escaping(self):
        # используем здесь мок, потому что ошибка
        # пишется в консоль, но не рейсится наружу
        fh = logging.StreamHandler()
        logging.root.handlers = []
        logging.root.addHandler(fh)
        fh.handleError = Mock()

        log("%0", level="ERROR")
        self.assertEqual(fh.handleError.called, False)

    def test_render_params(self):
        fh = logging.StreamHandler()
        logging.root.handlers = []
        logging.root.addHandler(fh)
        fh.stream.write = Mock()
        log("%(p)s %p %p", level="ERROR", params={'p': 'value'})
        self.assertEqual(fh.stream.write.call_args[0][0], 'value %p %p\n')



# coding: utf-8
import unittest

from smart_kit.text_preprocessing import remote


class TestTextAlt:
    def __init__(self, text):
        self.text = text

    def get_result(self, *argv, **kwargs):
        return self.text.lowercase()

    @classmethod
    def run(cls, x):
        return cls(x)


class TextPrerocessingTest1(unittest.TestCase):
    def setUp(self):
        self.test_text = 'Any text'
        self.test_context = ['Any context']
        self.test_message_type = 'Any message'

    def test_text_preprocessing_base_text_normalizer_init(self):
        obj1 = remote.BaseTextNormalizer(self.test_text)
        obj2 = remote.BaseTextNormalizer(self.test_text, '')
        obj3 = remote.BaseTextNormalizer(self.test_text, self.test_context, self.test_message_type)
        self.assertTrue(obj1.text == self.test_text)
        self.assertTrue(obj1.context == [])
        self.assertTrue(obj1.message_type == "text")
        self.assertTrue(obj2.text == self.test_text)
        self.assertTrue(obj2.context == [])
        self.assertTrue(obj2.message_type == "text")
        self.assertTrue(obj3.text == self.test_text)
        self.assertTrue(obj3.context == self.test_context)
        self.assertTrue(obj3.message_type == self.test_message_type)

    def test_text_preprocessing_base_text_normalizer_get_result(self):
        obj1 = remote.BaseTextNormalizer(self.test_text, self.test_context, self.test_message_type)
        with self.assertRaises(NotImplementedError):
            obj1.get_result(*(1, 2, 3), **{'first': 1, 'second': 2})
        with self.assertRaises(NotImplementedError):
            obj1.get_result()
        with self.assertRaises(NotImplementedError):
            obj1.get_result(*(1, 2, 3))
        with self.assertRaises(NotImplementedError):
            obj1.get_result(**{'first': 1, 'second': 2})

    def test_text_preprocessing_base_text_normalizer_normalize(self):
        obj1 = remote.BaseTextNormalizer(self.test_text, self.test_context, self.test_message_type)
        with self.assertRaises(NotImplementedError):
            self.assertTrue(obj1.normalize(self.test_text) == 'any text')
        with self.assertRaises(NotImplementedError):
            self.assertTrue(obj1.normalize(self.test_text,  *(1, 2)) == 'any text')
        with self.assertRaises(NotImplementedError):
            self.assertTrue(obj1.normalize(self.test_text, **{'first': 1, 'second': 2}) == 'any text')
        with self.assertRaises(NotImplementedError):
            self.assertTrue(obj1.normalize(self.test_text, *(1, 2), **{'first': 1, 'second': 2}) == 'any text')




# coding: utf-8
import unittest

from smart_kit.text_preprocessing import http_text_normalizer


class TextPrerocessingTest2(unittest.TestCase):
    def setUp(self):
        self.test_text = 'Any text'
        self.test_context = ['Any context']
        self.test_message_type = 'Any message'

    def test_normalization_words_tokenized_set(self):
        self.assertTrue(http_text_normalizer.words_tokenized_set('a s d f') == {'a', 's', 'd', 'f'})
        self.assertTrue(http_text_normalizer.words_tokenized_set(' ') == {'', '', ''})

    def test_normalization_print_current_state_init(self):
        obj1 = http_text_normalizer.HttpTextNormalizer("http://any-url", 111, 8, True)
        self.assertTrue(obj1.TEXT_PARAM_NAME == "text")
        self.assertTrue(obj1.NORMALIZE_METHOD == "normalize")
        self.assertTrue(obj1._preprocess_url == 'http://any-url/preprocess')
        self.assertTrue(obj1._classify_url == 'http://any-url/classify')
        self.assertTrue(obj1._normalize_url == 'http://any-url/normalize')
        self.assertTrue(obj1._timeout == 8)
        self.assertTrue(obj1._verbose)
        self.assertTrue(obj1._tqdm_func == http_text_normalizer.tqdm)

    def test_normalization_set_preprocess_mode(self):
        obj1 = http_text_normalizer.HttpTextNormalizer("http://any-url", 111, 8, True)
        obj2 = http_text_normalizer.HttpTextNormalizer("http://any-url", 111, 8, False)
        obj1.set_preprocess_mode()
        obj1.set_classify_mode()
        obj1.set_normalize_mode()
        obj2.set_preprocess_mode()
        obj2.set_classify_mode()
        obj2.set_normalize_mode()



KEY_NAME = "key_name"

STARTUP_VALUE = "app_startup"
RUNNING_VALUE = "app_running"
STATS_VALUE = "stats"
ANSWER_VALUE = "answer"
UPDATE_CONTENT_VALUE = "app_update_content"
SERVICE_MANAGEMENT_VALUE = "service_management"
SERVICE_MANAGEMENT_TIME_VALUE = "service_management_time_ms"
SKIP_INCOMING_MESSAGE_VALUE = "skip_incoming_message"
FAILED_DB_INTERACTION = "failed_db_interaction"

GREENFIELD_VALUE = "greenfiels"
AB_GROUPS_VALUE = "ab_groups"
REQUEST_VALUE = "request"
SERVICE_REQUEST_VALUE = "service_request"
REPLY_VALUE = "reply"
CHOSEN_SCENARIO_VALUE = "chosen_scenario"
CHOSEN_ACTION_VALUE = "chosen_action"
CHOSEN_SUB_SCENARIO_VALUE = "chosen_sub_scenario"
DISTANCE_VALUE = "distance"
SCENARIO_DESCRIPTION_VALUE = "scenario_description"
ROOT_ID_VALUE = "root_id"
INTENT_ID_VALUE = "intent_id"
FILLER_EXTRACT_VALUE = "filler_extract"
SCENARIO_RESULT_VALUE = "scenario_result"
REQUIREMENT_CHECK_VALUE = "requirement_check"
CHECKING_NODE_ID_VALUE = "checking_node_id"
NORMALIZED_TEXT_VALUE = "normalized_text"
PREPROCESSING_TIME_VALUE = "preprocessing_time_ms"
TIME_FROM_KAFKA_MSG_CREATION = "time_from_kafka_msg_creation"
ACTION_BEFORE_VALUE = "action_before"
DIALOG_ERROR_VALUE = "dialog_error"
FILLER_RESULT_VALUE = "filler_result"
CHOSEN_NODE_ID_VALUE = "chosen_node_id"
BEGIN_RECORD_VALUE = "BEGIN_RECORD"
END_RECORD_VALUE = "END_RECORD"
SCENARIO_VALUE = "scenario"
CLASSIFIER_VALUE = "classifier"
SELF_SERVICE_RUN_VALUE = "self_service_run"
SELF_SERVICE_MISSTATE_VALUE = "self_service_misstate"
SELF_SERVICE_SUCCESS_VALUE = "self_service_success"
SELF_SERVICE_FAIL_VALUE = "self_service_fail"

BEHAVIOR_ADD_VALUE = "behavior_add"
BEHAVIOR_MISSTATE_VALUE = "behavior_misstate"
BEHAVIOR_CHECK_MISSTATE_VALUE = "behavior_check_misstate"
BEHAVIOR_DATA_VALUE = "behavior_data"
BEHAVIOR_GOT_SAVED_VALUE = "behavior_got_saved"
BEHAVIOR_SUCCESS_VALUE = "behavior_success"
BEHAVIOR_FAIL_VALUE = "behavior_fail"
BEHAVIOR_TIMEOUT_VALUE = "behavior_timeout"
BEHAVIOR_CALLBACK_ID_VALUE = "behavior_callback_id"
BEHAVIOR_ID_VALUE = "behavior_id"
HOSTNAME = "hostname"
LAST_HOSTNAME = "last_hostname"
CHECK_HOSTNAME = "check_hostname"

NORMALIZE_INTENT_VALUE = "normalize_intent"
SKIPPED_INTENT_VALUE = 'skipped_intent'
INVALID_INTENT_VALUE = 'invalid_intent'
CONTAINER_NAME_VALUE = 'container_name'
CONTAINER_REQUIREMENT_CHECK_VALUE = 'container_requirement_check_value'


AB_GROUPS_MESSAGE = "%(key_name)s=%(ab_groups)s"
REPLY_MESSAGE = "%(key_name)s=%(reply)s"
LAST_SCENARIO_MESSAGE = "Dialog manager run chosen scenario info: %(key_name)s=%(chosen_scenario)s, " \
                        "scenario_description=%(scenario_description)s"
DIALOG_SCENARIO_MESSAGE = "Dialog manager run last scenario info: %(key_name)s=%(chosen_scenario)s, " \
                          "distance=%(distance)s, scenario_description=%(scenario_description)s, " \
                          "root_id=%(root_id)s"
INVALID_INTENT_MESSAGE = 'Invalid intent, intent=%(intent_id)s'
CLASSIFIER_MESSAGE = "classifier: %(classifier_name)s, result: %(result)s, " \
                     "weights: %(score)s, time: %(time)s ms"
CHOSEN_SCENARIO_MESSAGE = "%(key_name)s=%(chosen_scenario)s"
CHOSEN_ACTION_MESSAGE = "%(key_name)s=%(chosen_action)s"
CHOSEN_SUB_SCENARIO_MESSAGE = "%(key_name)s=%(chosen_sub_scenario)s"
DISTANCE_MESSAGE = "%(key_name)s=%(distance)s"
INTENT_ID_MESSAGE = "%(key_name)s=%(intent_id)s"
FILLER_EXTRACT_MESSAGE = "%(key_name)s=%(filler_extract)s"
SCENARIO_RESULT_MESSAGE = "%(key_name)s=%(scenario_result)s"
REQUIREMENT_CHECK_MESSAGE = "%(key_name)s=%(requirement_check)s"
CHECKING_NODE_ID_MESSAGE = "%(key_name)s=%(checking_node_id)s"
NORMALIZED_TEXT_MESSAGE = "%(key_name)s=%(normalized_text)s"
ACTION_BEFORE_MESSAGE = "%(key_name)s=%(action_before)s"
DIALOG_ERROR_MESSAGE = "%(key_name)s=%(dialog_error)s"
FILLER_RESULT_MESSAGE = "%(key_name)s=%(filler_result)s"
CHOSEN_NODE_ID_MESSAGE = "%(key_name)s=%(chosen_node_id)s"
SKIPPED_INTENT_MESSAGE = "Skipped intent %(intent_id)s, scenario is not available"
CONTAINER_REQUIREMENT_CHECK_MESSAGE = 'Classifier container requirement check: result: ' \
                                      '%(requirement_check)s, container: %(container_name)s'
MESSAGE_ID = "message_id"

CALLBACK_ID_HEADER = "app_callback_id"



from time import time
from unittest import TestCase

from scenarios.scenario_models.history import Event, History, HistoryEventFormatter
from smart_kit.utils.picklable_mock import PicklableMock


class ScenarioHistoryTest(TestCase):

    def test_event(self):
        item = {
            'scenario': 'name',
            'scenario_version': 'version',
            'node': 'node',
            'result': 'true',
            'type': 'type',
            'content': {'foo': 'bar'}
        }
        event = Event(**item)

        self.assertEqual(event.scenario, item['scenario'])
        self.assertEqual(event.scenario_version, item['scenario_version'])
        self.assertEqual(event.node, item['node'])
        self.assertEqual(event.result, item['result'])
        self.assertEqual(event.type, item['type'])
        self.assertDictEqual(event.content, item['content'])

    def test_event_to_dict(self):
        expected = {
            'scenario': 'name',
            'scenario_version': 'version',
            'node': 'node',
            'result': 'true',
            'type': 'type',
            'content': {'foo': 'bar'},
            'created_time': time()
        }
        event = Event(**expected)

        self.assertDictEqual(event.to_dict(), expected)

    def test_history_add_event(self):
        descriptions = PicklableMock()
        item = {
            'type': 'event_type',
            'content': {'foo': 'bar'}
        }
        history = History({}, descriptions, None)
        expected = Event(**item)

        history.add_event(expected)

        self.assertEqual(len(history.get_raw_events()), 1)
        self.assertEqual(history.get_raw_events()[0], expected)

    def test_history_clear(self):
        descriptions = PicklableMock()
        items = {
            'events': [
                {
                    'type': 'event_type_1',
                    'content': {'foo': 'bar'}
                },
                {
                    'type': 'event_type_2',
                    'content': {'foo': 'bar'}
                }
            ]
        }
        history = History(items, descriptions, None)
        self.assertEqual(len(history.get_raw_events()), 2)

        history.clear()
        self.assertEqual(len(history.get_raw_events()), 0)

    def test_history_raw(self):
        now = time()
        descriptions = PicklableMock()
        items = {
            'events': [
                {
                    'type': 'event_type_1',
                    'content': {'foo': 'bar'},
                    'created_time': now
                }
            ]
        }
        expected = {
            'events': [
                {
                    'type': 'event_type_1',
                    'content': {'foo': 'bar'},
                    'node': None,
                    'result': None,
                    'scenario': None,
                    'scenario_version': None,
                    'created_time': now
                }
            ]
        }

        history = History(items, descriptions, None)

        self.assertDictEqual(history.raw, expected)

    def test_history_expire(self):
        now = time()
        descriptions = PicklableMock()
        descriptions.event_expiration_delay = 5
        items = {
            'events': [
                {
                    'type': 'event_type_1',
                    'content': {'foo': 'bar'},
                    'created_time': now - 1
                },
                {
                    'type': 'event_type_2',
                    'content': {'foo': 'bar'},
                    'created_time': now - 5
                }
            ]
        }
        expected_keys = {'event_type_1',}

        history = History(items, descriptions, None)
        history.expire()

        history_raw = history.raw
        events_raw = history_raw["events"]
        event_keys = {event_raw.get("type") for event_raw in events_raw}

        self.assertSetEqual(event_keys, expected_keys)

    def test_history_event_formatter(self):
        events = [
            Event(type='field_event', scenario='name', node='node', result='filled', content={'field': 'foo'}),
            Event(type='field_event', scenario='name', node='node', result='filled', content={'field': 'bar'}),
        ]
        expected = [
            {
                'no': 1,
                'scenarioName': 'name',
                'scenarioVersion': None,
                'result': 'filled',
                'eventType': 'field_event',
                'eventContent': {'field': 'foo'}
            },
            {
                'no': 2,
                'scenarioName': 'name',
                'scenarioVersion': None,
                'result': 'filled',
                'eventType': 'field_event',
                'eventContent': {'field': 'bar'}
            }
        ]

        formatter = HistoryEventFormatter()

        self.assertListEqual(formatter.format(events), expected)

    def test_get_events(self):
        descriptions = PicklableMock()
        descriptions.formatter = HistoryEventFormatter()
        items = {
            'events': [
                {
                    'type': 'field_event',
                    'scenario': 'name',
                    'node': 'node',
                    'result': 'filled',
                    'content': {'field': 'field_name'}
                }
            ]
        }
        expected = [
            {
                'no': 1,
                'scenarioName': 'name',
                'scenarioVersion': None,
                'result': 'filled',
                'eventType': 'field_event',
                'eventContent': {'field': 'field_name'}
            }
        ]

        history = History(items, descriptions, None)
        events = history.get_events()

        self.assertDictEqual(events[0], expected[0])



from core.db_adapter import error
from core.db_adapter.db_adapter import DBAdapter


class MemoryAdapter(DBAdapter):

    def __init__(self, config=None):
        super(DBAdapter, self).__init__(config)
        self.memory_storage = {}

    def _glob(self, path, pattern):
        raise error.NotSupportedOperation

    def _path_exists(self, path):
        raise error.NotSupportedOperation

    def _on_prepare(self):
        pass

    def connect(self):
        pass

    def _open(self, filename, *args, **kwargs):
        pass

    def _save(self, id, data):
        self.memory_storage[id] = data

    def _replace_if_equals(self, id, sample, data):
        stored_data = self.memory_storage.get(id)
        if stored_data == sample:
            self.memory_storage[id] = data
            return True
        return False

    def _get(self, id):
        data = self.memory_storage.get(id)
        return data

    def _list_dir(self, path):
        pass



from smart_kit.management.base import AppCommand
import os

from core.repositories.file_repository import FileRepository

import json


class CreateCacheCommand(AppCommand):
    """Create cache for normalization results by parsing static"""

    def __init__(self, app_config):
        self._normalize_key = "symbols_for_normalize"
        self._ext = ".json"
        self.app_config = app_config
        self.recognizer_client = app_config.NORMALIZER.with_cache(
            app_config,
            cache_lifetime=float("+inf"),
            verbose=True
        )
        self.recognizer_client.cache.clear()

    def execute(self, *args, **kwargs):
        self.recognizer_client.normalize_sequence(self.items_for_normalized)
        self.recognizer_client.cache.save(os.path.join(self.app_config.REFERENCES_PATH, ".normalization_cache.json"))

    def _collect_items_for_normalized(self, data):
        if hasattr(data, 'items'):
            for key, val in data.items():
                if key == self._normalize_key:
                    yield val
                if isinstance(val, dict):
                    for result in self._collect_items_for_normalized(val):
                        yield result
                elif isinstance(val, list):
                    for item in val:
                        for result in self._collect_items_for_normalized(item):
                            yield result

    @property
    def items_for_normalized(self):
        references_path = self.app_config.REFERENCES_PATH
        items_for_normalized = []
        for r, d, f in os.walk(references_path):
            for file in f:
                if self._ext in file:
                    filename = os.path.join(r, file)
                    rep = FileRepository(filename=filename, loader=json.loads)
                    rep.load()
                    data = rep.data
                    item = self._split_items(self._collect_items_for_normalized(data))
                    items_for_normalized.extend(item)
        return items_for_normalized

    def _split_items(self, items_for_normalized):
        result = []
        for item in items_for_normalized:
            if isinstance(item, dict):
                for key, val in item.items():
                    if isinstance(val, list):
                        result.extend(val)
                    else:
                        result.append(val)
            elif isinstance(item, list):
                result.extend(item)
            else:
                result.append(item)
        return result



import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring


class Rerunable():
    DEFAULT_RERUNABLE_TRY_COUNT = 5

    def __init__(self, config=None):
        self.config = config or {}
        self.try_count = self.config.get("try_count") or self.DEFAULT_RERUNABLE_TRY_COUNT

    @property
    def _handled_exception(self):
        raise NotImplementedError

    def _on_prepare(self):
        raise NotImplementedError

    def _on_all_tries_fail(self):
        raise NotImplementedError

    def _run(self, action, *args, _try_count=None, **kwargs):
        if _try_count is None:
            _try_count = self.try_count
        if _try_count <= 0:
            self._on_all_tries_fail()
        _try_count = _try_count - 1
        try:
            result = action(*args, **kwargs)
        except self._handled_exception as e:
            params = {
                "class_name": str(self.__class__),
                "exception": str(e),
                "try_count": _try_count,
                log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE
            }
            log("%(class_name)s run failed with %(exception)s.\n Got %(try_count)s tries left.",
                          params=params,
                          level="ERROR")
            self._on_prepare()
            result = self._run(action, *args, _try_count=_try_count, **kwargs)
            counter_name = self._get_counter_name()
            if counter_name:
                monitoring.got_counter(f"{counter_name}_exception")
        return result

    def _get_counter_name(self):
        return



import requests
import json


class BaseTextNormalizer:
    def __init__(self, text, context=None, message_type="text"):
        self.text = text
        self.context = context or []
        self.message_type = message_type

    @classmethod
    def normalize(cls, original_text, *args, **kwargs):
        inst = cls(text=original_text)
        return inst.get_result(*args, **kwargs)

    def get_result(self, *args, **kwargs):
        raise NotImplementedError


class HTTPTextNormalizer(BaseTextNormalizer):

    def get_result(self, address):
        response = self._get_response(address)
        return json.loads(response.text, encoding=response.encoding)

    def _get_response(self, address):
        response = requests.get(address, params={"original_text": self.text})
        response.raise_for_status()
        return response



# coding: utf-8
import unittest
from unittest.mock import Mock

from scenarios.user import parametrizer
from smart_kit.utils.picklable_mock import PicklableMock


class TestMessage:
    payload = "any payload"
    uuid = "1234-5678-9102"


class UserTest1(unittest.TestCase):
    def setUp(self):
        self.test_user1 = Mock('User')
        self.test_user1.message = Mock('Message')
        self.test_user1.descriptions = {"scenarios": {11: 'any data 1', 12: 'any data 2'}}
        self.test_user1.last_scenarios = Mock('last scenario')
        self.test_user1.last_scenarios.last_scenario_name = 11
        self.test_items = {"123-ASDF": {"command": [], "no_empty_nodes": True}}
        self.test_scenario1 = Mock('scenario')
        self.test_scenario1.scenario_description = "This is test scenario 1 desc"
        self.test_scenario1.form_type = "any type 1"
        self.test_user2 = Mock('User')
        self.test_user2.message = Mock('Message')
        self.test_user2.descriptions = {"scenarios": {22: self.test_scenario1}}
        self.test_user2.last_scenarios = Mock('last scenario')
        self.test_user2.last_scenarios.last_scenario_name = 22
        self.test_user3 = Mock('User')
        self.test_user3.message = Mock('Message')
        self.test_user3.descriptions = {"scenarios": {22: self.test_scenario1}}
        self.test_user3.last_scenarios = Mock('last scenario')
        self.test_user3.last_scenarios.last_scenario_name = 33
        self.test_forms = {'any type 1': 'any form 1', 'any type 2': 'any form 2'}
        self.test_user4 = Mock('User')
        self.test_user4.forms = Mock('Forms')
        self.test_user4.forms.collect_form_fields = lambda: self.test_forms
        self.test_user4.message = TestMessage()
        self.test_user4.message.payload = "any payload"
        self.test_user4.message.uuid = "1234-5678-9102"
        self.test_user4.variables = Mock('Variables')
        self.test_user4.variables.values = [1, 2, 3]
        self.test_user4.counters = Mock('Counters')
        self.test_user4.counters.raw = 3
        self.test_user4.descriptions = {"scenarios": {22: self.test_scenario1}}
        self.test_user4.last_scenarios = Mock('last scenario')
        self.test_user4.last_scenarios.last_scenario_name = 22
        self.test_user4.gender_selector = PicklableMock()
        self.test_user4.gender_selector.get_text_by_key = ''
        self.test_user4.settings = {}
        local_vars = PicklableMock()
        local_vars.values = dict()
        self.test_user4.local_vars = local_vars
        self.test_text_preprocessing_result = Mock('text_preprocessing_result')
        self.test_text_preprocessing_result.raw = {1: "any text"}

    def test_parametrizer_SmartAppParametrize_init(self):
        obj = parametrizer.Parametrizer(self.test_user1, self.test_items)
        self.assertTrue(obj._user == self.test_user1)

    def test_parametrizer_SmartAppParametrize_get_scenario(self):
        obj = parametrizer.Parametrizer(self.test_user1, self.test_items)
        self.assertTrue(obj._get_scenario() == 'any data 1')

    def test_parametrizer_get_main_form(self):
        obj1 = parametrizer.Parametrizer(self.test_user2, self.test_items)
        obj2 = parametrizer.Parametrizer(self.test_user3, self.test_items)
        self.assertTrue(obj1._get_main_form(self.test_forms) == 'any form 1')
        self.assertIsNone(obj2._get_main_form(self.test_forms))

    def test_parametrizer_get_user_data(self):
        obj1 = parametrizer.Parametrizer(self.test_user4, self.test_items)
        answer1 = {'payload': 'any payload', 'uuid': '1234-5678-9102', 'forms': {
            'any type 1': 'any form 1', 'any type 2': 'any form 2'}, 'main_form': 'any form 1',
                   'text_preprocessing_result': {}, 'variables': [1, 2, 3], 'counters': 3, 'scenario_id': 22,
                   'gender_sensitive_text': '', 'local_vars': {}, 'settings': {}}
        answer2 = {'payload': 'any payload', 'uuid': '1234-5678-9102', 'forms': {
            'any type 1': 'any form 1', 'any type 2': 'any form 2'}, 'main_form': 'any form 1',
                   'text_preprocessing_result': {1: 'any text'}, 'variables': [1, 2, 3], 'counters': 3,
                   'scenario_id': 22, 'gender_sensitive_text': '', 'local_vars': {}, 'settings': {}}
        result1 = obj1._get_user_data()
        result1.pop('message')
        result2 = obj1._get_user_data(self.test_text_preprocessing_result)
        result2.pop('message')
        self.assertTrue(result1 == answer1)
        self.assertTrue(result2 == answer2)



import unittest

from core.descriptions.descriptions import Descriptions, registered_description_factories, default_description_factory
from core.repositories.base_repository import BaseRepository


class MockDescription1:
    def __init__(self, raw_data):
        self.raw_data = raw_data

    @property
    def raw(self):
        return self.raw_data


class MockDescription2(MockDescription1):
    pass


class MockRepository(BaseRepository):
    def __init__(self, data, key):
        super(MockRepository, self).__init__(key=key)
        self.data = data


class DescriptionsTest(unittest.TestCase):
    def setUp(self):
        val1 = MockRepository("raw_1", "repo_key1")
        val2 = MockRepository("raw_2", "repo_key2")
        val3 = MockRepository("raw_3", "repo_key3")
        self.registered_repos = {"repo_key1": val1, "repo_key2": val2, "repo_key3": val3}
        registered_description_factories[self.registered_repos["repo_key1"].key] = MockDescription1
        registered_description_factories[self.registered_repos["repo_key2"].key] = MockDescription2
        registered_description_factories[self.registered_repos["repo_key3"].key] = default_description_factory
        self.descriptions = Descriptions(registered_repositories=self.registered_repos)

    def test_get1(self):
        item1 = self.descriptions["repo_key1"]
        assert item1.raw == "raw_1"
        assert type(item1) == MockDescription1

    def test_get2(self):
        item2 = self.descriptions["repo_key2"]
        assert item2.raw == "raw_2"
        assert type(item2) == MockDescription2

    def test_get3(self):
        item2 = self.descriptions["repo_key3"]
        assert item2 == "raw_3"



# coding: utf-8
import unittest
from unittest.mock import Mock

from core.basic_models.actions.command import Command
from core.basic_models.actions.string_actions import StringAction
from smart_kit.names.message_names import NOTHING_FOUND
from smart_kit.system_answers import nothing_found_action


class SystemAnswersTest1(unittest.TestCase):
    def setUp(self):
        self.test_command_1 = Mock('Command')
        self.test_id = '123-345-678'  # пусть чему-то равняется
        self.test_items1 = {"123-ASDF": {"command": [], "no_empty_nodes": True}}
        self.test_text_preprocessing_result = Mock('text_preprocessing_result')
        self.test_user1 = Mock('user')
        self.test_user1.parametrizer = Mock('parametrizer')
        self.test_user1.parametrizer.collect = lambda x, filter_params: {'input': x, 'result': x}

    def test_system_answers_nothing_found_action_init(self):
        obj1 = nothing_found_action.NothingFoundAction()
        obj2 = nothing_found_action.NothingFoundAction(self.test_items1, self.test_id)
        self.assertIsNone(obj1.id)
        self.assertTrue(obj2.id == self.test_id)
        self.assertTrue(isinstance(obj1._action, StringAction))
        self.assertTrue(obj1._action.command == NOTHING_FOUND)

    def test_system_answer_nothing_found_action_run(self):
        obj1 = nothing_found_action.NothingFoundAction()
        obj2 = nothing_found_action.NothingFoundAction(self.test_items1, self.test_id)
        self.assertTrue(isinstance(obj1.run(self.test_user1, self.test_text_preprocessing_result).pop(), Command))
        self.assertTrue(isinstance(obj2.run(self.test_user1, self.test_text_preprocessing_result).pop(), Command))



# coding: utf-8
from scenarios.scenario_models.field.field_descriptions.basic_field_description import BasicFieldDescription


class IntegrationFieldDescription(BasicFieldDescription):

    def __init__(self, items, id):
        super(IntegrationFieldDescription, self).__init__(items, id)
        self._retry_count = items.get("retry_count", 0)



from typing import Dict, Any, Optional

from core.basic_models.requirement.basic_requirements import Requirement, requirement_factory
from core.model.base_user import BaseUser

from core.text_preprocessing.base import BaseTextPreprocessingResult

from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems


class ExternalRequirements(SmartUpdatableDescriptionsItems):
    def __init__(self, items):
        super(ExternalRequirements, self).__init__(requirement_factory, items, ordered=True)


class ExternalRequirement(Requirement):
    requirement: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ExternalRequirement, self).__init__(items, id)
        self.requirement = items["requirement"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        requirement = user.descriptions["external_requirements"][self.requirement]
        return requirement.check(text_preprocessing_result, user, params)



from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.field_filler_description import AvailableInfoFiller
from smart_kit.utils.picklable_mock import PicklableMock


class MockParametrizer:
    def __init__(self, user, items=None):
        self.items = items or {}
        self.user = user
        self.filter = items.get("filter") or False

    def collect(self, text_preprocessing_result=None, filter_params=None):
        data = {
            "person_info": self.user.person_info.raw,
            "payload": self.user.message.payload,
            "uuid": self.user.message.uuid

        }
        if self.filter:
            data.update({"filter": "filter_out"})
        return data


class TestAvailableInfoFiller(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.address = "Address!"
        payload_items = {'value': '{{payload.sf_answer.address}}'}
        cls.payload_filler = AvailableInfoFiller(payload_items)

    def setUp(self):
        template = PicklableMock()
        template.get_template = Mock(return_value=[])
        user = PicklableMock()
        user.parametrizer = MockParametrizer(user, {})
        user.message = PicklableMock()
        user.person_info = PicklableMock()
        user.descriptions = {"render_templates": template}
        self.user = user

    def test_getting_person_info_value(self):
        name = "Name!"
        surname = "Surname!"
        self.user.person_info.raw = PicklableMock()
        self.user.person_info.raw.full_name = {"name": name, "surname": surname}
        person_info_items = {'value': '{{person_info.full_name.surname}}'}
        person_info_filler = AvailableInfoFiller(person_info_items)

        result = person_info_filler.extract(None, self.user)
        self.assertEqual(result, surname)

    def test_getting_payload_value(self):
        self.user.message.payload = {"sf_answer": {"address": self.address}}
        result = self.payload_filler.extract(None, self.user)
        self.assertEqual(result, self.address)

    def test_getting_uuid_value(self):
        uuid = "15"
        self.user.message.uuid = {"chatId": uuid}
        uuid_items = {'value': '{{uuid.chatId}}'}
        uuid_filler = AvailableInfoFiller(uuid_items)

        result = uuid_filler.extract(None, self.user)
        self.assertEqual(result, uuid)

    def test_not_failing_on_wrong_path(self):
        self.user.message.payload = {"other_answer": {"address": self.address}}
        result = self.payload_filler.extract(None, self.user)
        self.assertIsNone(result)

    def test_return_empty_value(self):
        self.user.message.payload = {"sf_answer": '1'}
        result = self.payload_filler.extract(None, self.user)
        self.assertEqual("", result)

    def test_filter(self):
        template = PicklableMock()
        template.get_template = Mock(return_value=["payload.personInfo.identityCard"])
        self.user.parametrizer = MockParametrizer(self.user, {"filter": True})
        self.user.message.payload = {"personInfo": {"identityCard": "my_pass"}}
        self.user.descriptions = {"render_templates": template}
        payload_items = {'value': '{{filter}}'}
        filler = AvailableInfoFiller(payload_items)
        result = filler.extract(None, self.user)
        self.assertEqual("filter_out", result)

    def test_getting_payload_parsed_value(self):
        data = [
            {
                "id": 1,
                "surname": "Иванов",
                "name": "Иван",
                "patronimic": "Иванович",
                "tn": "1258838",
                "short_oe": "Д/О 8623/0302",
                "id_oe": "12344321",
                "count_tmc": 10,
                "actual_mol": True,
                "email": "user@omega.sbrf.ru",
                "uname": "IVANOV-II1",
                "position": "инженер"
            },
            {
                "id": 2,
                "surname": "Сидоров",
                "name": "Сидр",
                "patronimic": "Сидорович",
                "tn": "1258838",
                "short_oe": "Д/О 8623/0302",
                "id_oe": "12345678",
                "count_tmc": 5,
                "actual_mol": True,
                "email": "user@omega.sbrf.ru",
                "uname": "IVANOV-II1",
                "position": "инженер"
            },
            {
                "id": 3,
                "surname": "Петров",
                "name": "Петр",
                "patronimic": "Петрович",
                "tn": "144264",
                "short_oe": "Д/О 8623/030233",
                "id_oe": "12345678",
                "count_tmc": 5,
                "actual_mol": True,
                "email": "user@omega.sbrf.ru",
                "uname": "IVANOV-II1",
                "position": "инженер"
            }
        ]
        payload_items = {'value': '{{payload.data|tojson}}',
                         'loader': 'json'}
        payload_filler = AvailableInfoFiller(payload_items)

        self.user.message.payload = {
            "error_code": 200,
            "error_text": "OK",
            "skip": 1,
            "top": 20,
            "total_count": 125,
            "cacheGuid": "FHGDDASDHDAKSGFLAK",
            "data": data
        }
        result = payload_filler.extract(None, self.user)
        self.assertEqual(result, data)



# coding: utf-8
from typing import Dict

from core.basic_models.counter.counter import Counter


class Counters:

    def __init__(self, items: Dict[str, Dict], user):
        self._raw_items = items or {}
        self._items: Dict[str, Counter] = {}
        self._item_type = Counter

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key):
        counter = self._items.get(key)
        if counter is None:
            raw_data = self._raw_items.get(key)
            counter = self._item_type(raw_data)
            self._items[key] = counter
        return counter

    def clear(self, key):
        if key in self._items:
            del self._items[key]
        if key in self._raw_items:
            self._raw_items.pop(key)

    def expire(self):
        items = list(self.raw.keys())
        for key in items:
            if self[key].check_expire():
                self.clear(key)

    @property
    def raw(self):
        for name in self._items:
            counter = self._items[name]
            value = counter.raw
            if value:
                self._raw_items[name] = value
            else:
                if name in self._raw_items:
                    self._raw_items.pop(name)
        return self._raw_items



import unittest

from core.basic_models.answer_items.answer_items import BubbleText, ItemCard, SuggestDeepLink, SuggestText, \
    PronounceText, RawItem


class AnswerItemTest(unittest.TestCase):

    def test_ItemCard(self):
        nodes = {
            "ans": 42
        }
        items = {
            "text": "ans"
        }
        test = ItemCard(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'card': 42})

    def test_SuggestDeepLink(self):
        nodes = {
            "ans": 42,
            "dl": 24
        }
        items = {
            "deep_link": "dl",
            "title": "ans"
        }
        test = SuggestDeepLink(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'title': 42, 'action': {'deep_link': 24, 'type': 'deep_link'}})

    def test_SuggestText(self):
        nodes = {
            "ans": 42,
            "text": "text"
        }
        items = {
            "text": "text",
            "title": "ans"
        }
        test = SuggestText(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'title': 42, 'action': {'text': 'text', 'type': 'text'}})

    def test_SuggestText_no_nodes(self):
        items = {
            "text": "text",
            "title": "ans"
        }
        test = SuggestText(items)
        res = test.render({})
        self.assertDictEqual(res, {'title': "ans", 'action': {'text': 'text', 'type': 'text'}})

    def test_PronounceText(self):
        nodes = {
            "ans": 42
        }
        items = {
            "text": "ans"
        }
        test = PronounceText(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'pronounceText': 42})


class BubbleItemTest(unittest.TestCase):

    def test_BubbleText(self):
        nodes = {
            "ans": 42
        }
        items = {
            "text": "ans"
        }
        test = BubbleText(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'bubble': {'text': 42, "markdown": True}})

    def test_Bubble_miss_text(self):
        items = {
            "tsext": "ans"
        }
        self.assertRaises(KeyError, BubbleText, items)

    def test_BubbleText(self):
        nodes = {
            "ans": 42
        }
        items = {
            "text": "ans",
            "markdown": False
        }
        test = BubbleText(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'bubble': {'text': 42, "markdown": False}})


class RawItemTest(unittest.TestCase):

    def test_finished(self):
        items = {
            "type": "raw",
            "key": "finished",
            "value": "True"
        }
        test = RawItem(items)
        res = test.render({})
        self.assertDictEqual(res, {'finished': "True"})

    def test_Raw_miss_key(self):
        items = {
            "text": "ans"
        }
        self.assertRaises(KeyError, RawItem, items)

    def test_Raw_sophisticated(self):
        nodes = {
            "ans": {
                "emotionId": "oups",
                "ttsAnimation": "sad"
            }
        }
        items = {
            "key": "emotion",
            "value": "ans"
        }
        test = RawItem(items)
        res = test.render(nodes)
        self.assertDictEqual(res, {'emotion': {'emotionId': 'oups', 'ttsAnimation': 'sad'}})


from unittest import TestCase

from scenarios.scenario_models.field.field_filler_description import FirstNumberFiller, \
    FirstCurrencyFiller
from smart_kit.utils.picklable_mock import PicklableMock


class TestFirstNumberFiller(TestCase):
    def test_1(self):
        expected = "5"
        items = {}
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.num_token_values = [expected]

        filler = FirstNumberFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)

    def test_2(self):
        items = {}
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.num_token_values = []

        filler = FirstNumberFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)


class TestFirstCurrencyFiller(TestCase):
    def test_1(self):
        expected = "ru"
        items = {}
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.ccy_token_values = [expected]

        filler = FirstCurrencyFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertEqual(expected, result)

    def test_2(self):
        items = {}
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.ccy_token_values = []

        filler = FirstCurrencyFiller(items)
        result = filler.extract(text_preprocessing_result, None)

        self.assertIsNone(result)



# coding: utf-8

from core.db_adapter.os_adapter import OSAdapter
from core.logging.logger_utils import log
import core.logging.logger_constants as log_const


class BaseRepository:
    def __init__(self, source=None, key=None):
        self.source = source or OSAdapter(None)
        self._data = None
        self.key = key

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    def load(self):
        params = {
            "repository_class_name": self.__class__.__name__,
            "repository_key": self.key,
            log_const.KEY_NAME: log_const.REPOSITORY_LOAD_VALUE
        }
        log("%(repository_class_name)s.load  %(repository_key)s repo loading completed.", params=params,
                   level="WARNING")

    def fill(self, data):
        self.data = data

    def clear(self):
        self.data.clear()
        log("%(repository_class_name)s.clear %(repository_key)s cleared.",
                      params={"repository_class_name": self.__class__.__name__,
                              "repository_key": self.key,
                              log_const.KEY_NAME: log_const.REPOSITORY_CLEAR_VALUE},
                      level="WARNING")

    def save(self, save_parameters):
        raise NotImplementedError

    def check_load_in_parts(self):
        return False



import random
from time import time


class DelayRunner:
    def __init__(self, max_delay_minutes):
        self.max_update_delay = max_delay_minutes * 60
        self._ts = 0
        self._run_item = None
        self._run_args = None

    def schedule_run(self, run_item, run_args):
        self._ts = time() + random.randint(0, self.max_update_delay)
        self._run_item = run_item
        self._run_args = run_args

    def check_can_run(self):
        return self._ts != 0 and time() >= self._ts

    def run(self):
        self._run_item(*self._run_args)
        self._clean()

    def _clean(self):
        self._ts = 0
        self._run_item = None
        self._run_args = None




import unittest

from scenarios.user.preprocessing_messages.preprocessing_messages_description import \
    PreprocessingMessagesDescription


class PreprocessingMessagesDescriptionTest(unittest.TestCase):

    def test_description(self):
        items = {"messages": 5, "lifetime": 600}
        description = PreprocessingMessagesDescription(items)
        self.assertEqual(description.max_message_count, 5)
        self.assertEqual(description.lifetime, 600)



import functools
from typing import Optional


def fallback_if_none(fallback_result_value=None, fallback_result_attr: Optional[str] = None):
    """ Parametrized function decorator that returns value if original function has returned None.
    Fallback value could be specified explicitly or specified by attribute name (for class methods).
    If fallback value cannot be resolved - an exception is thrown.

    :param fallback_result_value: Explicit fallback value to return
    :param fallback_result_attr: Name of the attribute of the object that method assigned to
    :return: decorator
    """
    def _decorator(fn):
        @functools.wraps(fn)
        def _wrapper(*args, **kwargs):
            result = fn(*args, **kwargs)
            if result is not None:
                return result
            if fallback_result_value is not None:
                return fallback_result_value
            if fallback_result_attr is not None:
                fallback_result = getattr(args[0], fallback_result_attr)
                if fallback_result is not None:
                    return fallback_result
            raise ValueError(f'Function {fn.__module__}:{fn.__name__} has returned None, '
                             f'but fallback value or fallback attribute is not specified. Check decorator usage')
        return _wrapper
    return _decorator



# coding: utf-8
from collections import deque

from core.model.lazy_items import LazyItems


class LimitedQueuedHashableObjects:
    def __init__(self, items, description, user=None):
        self.description = description
        self.queue = deque(items or [], maxlen=description.max_len)

    def get_list(self):
        return list(self.queue)

    def add(self, id):
        self.queue.append(id)

    def clear(self):
        self.queue.clear()

    def check(self, id):
        duplicated = id in self.queue
        if not duplicated:
            self.add(id)
        return duplicated

    @property
    def raw(self):
        return self.get_list()


class LimitedQueuedHashableObjectsItems(LazyItems):
    def __init__(self, items, descriptions, user):
        super(LimitedQueuedHashableObjectsItems, self).__init__(items, descriptions, user, LimitedQueuedHashableObjects)



from time import time
from typing import Dict


class Counter:

    def __init__(self, items):
        items = items or {}
        self.create_time = items.get("create_time") or int(time())
        self.update_time = items.get("update_time")
        self.lifetime = items.get("lifetime")
        self.value = items.get("value")

    def check_expire(self) -> bool:
        return time() - self.create_time > self.lifetime if self.lifetime else False

    def inc(self, value=1, lifetime=None):
        self._update(value, lifetime)

    def dec(self, value=-1, lifetime=None):
        self._update(value, lifetime)

    def _update(self, value, lifetime):
        self.value = self.value or 0
        self.value += value
        self.update_time = int(time())
        if self.lifetime is None:
            self.lifetime = lifetime

    def set(self, value=None, reset_time=False, time_shift=0):
        new_value = value if value is not None else self.value
        self._set(new_value, reset_time, time_shift)

    def _set(self, value, reset_time=False, time_shift=0):
        self.value = value
        self.update_time = int(time()) + time_shift
        if reset_time:
            self.create_time = self.update_time

    def __eq__(self, amount):
        value = self._get_eq_value(amount)
        return value == amount

    def __ne__(self, amount):
        value = self._get_eq_value(amount)
        return value != amount

    def __lt__(self, amount):
        value = self._get_eq_value(amount)
        return value < amount

    def __gt__(self, amount):
        value = self._get_eq_value(amount)
        return value > amount

    def __le__(self, amount):
        value = self._get_eq_value(amount)
        return value <= amount

    def __ge__(self, amount):
        value = self._get_eq_value(amount)
        return value >= amount

    def _get_eq_value(self, amount):
        if amount is None:
            value = self.value
        else:
            value = self.value or 0
        return value

    @property
    def raw(self) -> Dict[str, int]:
        if self.value is not None:
            return {
                "value": self.value,
                "create_time": self.create_time,
                "update_time": self.update_time,
                "lifetime": self.lifetime
            }



# coding: utf-8
from typing import Iterable, Dict, Any, Optional

from core.model.factory import build_factory
from core.model.registered import Registered

comparators = Registered()

comparator_factory = build_factory(comparators)


class Comparator:
    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        pass

    def compare(self, left: Any, right: Any) -> bool:
        raise NotImplementedError


class MoreComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left > right


class LessComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left < right


class MoreOrEqualComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left >= right


class LessOrEqualComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left <= right


class EqualComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left == right


class NotEqualComparator(Comparator):
    def compare(self, left: int, right: int) -> bool:
        return left != right


class InComparator(Comparator):
    def compare(self, left: Any, right: Iterable[Any]) -> bool:
        return left in right



from functools import cached_property
from typing import TypeVar, Dict, Any

from core.basic_models.parametrizers.filter import Filter

T = TypeVar("T")


class BasicParametrizer:
    def __init__(self,  user, items):
        self._user = user
        self._filter = items.get("filter") or {}

    @cached_property
    def filter(self):
        return Filter(self._filter)

    def filter_out(self, data: T, filter_params=None) -> T:
        return self.filter.filter_out(data, self._user, filter_params)

    def _get_user_data(self, text_preprocessing_result=None) -> Dict[str, Any]:
        return {"message": self._user.message}

    def collect(self, text_preprocessing_result=None, filter_params=None) -> Dict[str, Any]:
        data = self._get_user_data(text_preprocessing_result)
        return self.filter_out(data, filter_params)



from functools import cached_property
from typing import Dict


class Device:
    def __init__(self, value):
        self._value = value

    @cached_property
    def value(self):
        return self._value

    @cached_property
    def platform_type(self) -> str:
        return self._value.get("platformType") or ""

    @cached_property
    def platform_version(self) -> str:
        return self._value.get("platformVersion") or ""

    @cached_property
    def surface(self) -> str:
        return self._value.get("surface") or ""

    @cached_property
    def surface_version(self) -> str:
        return self._value.get("surfaceVersion") or ""

    @cached_property
    def features(self) -> Dict:
        return self._value.get("features") or {}

    @cached_property
    def capabilities(self) -> Dict:
        return self._value.get("capabilities") or {}

    @cached_property
    def additional_info(self) -> Dict:
        return self._value.get("additionalInfo") or {}

    @cached_property
    def tenant(self) -> str:
        return self._value.get("tenant") or ""

    @cached_property
    def device_id(self) -> str:
        return self._value.get("deviceId") or ""

    @cached_property
    def ram_gb(self) -> str:
        return self._value.get("ramGb") or ""

    @cached_property
    def cpu(self) -> str:
        return self._value.get("cpu") or ""



# coding: utf-8
from functools import cached_property

from core.basic_models.requirement.basic_requirements import Requirement
from core.model.factory import factory


class LastScenariosDescription:
    def __init__(self, items, id):
        self.id = id
        self._channels = items.get("channels")
        self._requirement = items.get("requirement")
        self.count = items.get("count", 1)

    @cached_property
    @factory(Requirement)
    def requirement(self):
        return self._requirement

    def check(self, text_preprocessing_result, user):
        return user.message.channel in self._channels and self.requirement.check(text_preprocessing_result, user) if \
            self._channels else self.requirement.check(text_preprocessing_result, user)



from unittest import TestCase
from unittest.mock import Mock, MagicMock

from core.basic_models.actions.basic_actions import Action, action_factory, actions
from core.basic_models.actions.command import Command
from core.model.registered import registered_factories
from scenarios.scenario_descriptions.tree_scenario.tree_scenario import TreeScenario
from smart_kit.utils.picklable_mock import PicklableMock, PicklableMagicMock


class MockAction:
    def __init__(self, items=None, command_name=None):
        self.called = False
        self.command_name = command_name

    def run(self, user, text_preprocessing_result, params):
        self.called = True
        if self.command_name:
            return [Command(self.command_name)]


class BreakAction:
    def __init__(self, items=None):
        pass

    def run(self, user, text_preprocessing_result, params):
        user.scenario_models["some_id"].break_scenario = True
        return []


class TestTreeScenario(TestCase):

    def test_1(self):
        """
        Тест проверяет сценарий из одного узла. Предполагается идеальный случай, когда одно поле
        и мы смогли его заполнить.
        Тест очень сложен с точки зрения инициализации моков. Но это лучше чем усложнять в целом процесс тестирования,
        используя полноценные сущности.
        """

        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        actions["external"] = MockAction

        form_type = "form for doing smth"
        internal_form_key = "my form key"

        node_mock = {"form_key": internal_form_key, "actions": [{"type": "test", "command": "jump", "nodes": {}}]}

        items = {"channels": "web", "form": form_type, "start_node_key": "node_1",
                 "scenario_nodes": {"node_1": node_mock}}

        field_descriptor = PicklableMock(name="field_descriptor_mock")
        field_descriptor.filler.extract = PicklableMock(name="my_field_value_1", return_value=61)
        field_descriptor.fill_other = False
        field_descriptor.field_validator.actions = []

        internal_form = PicklableMock(name="internal_form_mock")
        internal_form.description.fields.items = PicklableMock(return_value=[("age", field_descriptor)])
        internal_form.field.field_validator.requirement.check = PicklableMock(return_value=True)
        internal_form.fields = PicklableMagicMock()
        internal_form.fields.values.items = PicklableMock(return_value={"age": 61})
        internal_form.is_valid = PicklableMock(return_value=True)

        start_form_mock = PicklableMagicMock(name="start_form_mock")

        composite_form = PicklableMagicMock(name="form_mock")
        composite_form.forms = start_form_mock
        composite_form.set_valid = PicklableMock()

        context_forms = PicklableMagicMock(name="context_form_mock")
        context_forms.new = PicklableMock(return_value=PicklableMock(forms={"my form key": internal_form},
                                                                     is_endless=PicklableMock(return_value=False)))

        current_node_mock = PicklableMock(name="current_node_mock")
        current_node_mock.form_key = "start_node_mock"
        current_node_mock.available_nodes = False

        user = PicklableMock()
        user.forms = context_forms
        user.scenario_models = PicklableMagicMock()
        user.scenario_models.__getitem__ = PicklableMock(name="scenario_mock", return_value=current_node_mock)

        text_preprocessing_result = None

        scenario = TreeScenario(items, 1)

        result = scenario.run(text_preprocessing_result, user)
        self.assertIsNone(current_node_mock.current_node)
        context_forms.new.assert_called_once_with(form_type)

    def test_breake(self):
        """
        Тест проверяет выход из сценария если сработает флаг break_scenario
        """

        registered_factories[Action] = action_factory
        actions["test"] = MockAction
        actions["break"] = MockAction
        actions["success"] = MockAction

        form_type = "form for doing smth"
        internal_form_key = "my form key"

        node_mock = {"form_key": internal_form_key, "actions": [{"type": "test", "command": "jump", "nodes": {}}]}

        items = {"channels": "web", "form": form_type, "start_node_key": "node_1",
                 "scenario_nodes": {"node_1": node_mock}, "actions": [{"type": "success"}]}

        field_descriptor = PicklableMock(name="field_descriptor_mock")
        field_descriptor.filler.extract = PicklableMock(name="my_field_value_1", return_value=61)
        field_descriptor.fill_other = False
        field_descriptor.field_validator.actions = []
        field_descriptor.on_filled_actions = [BreakAction(), MockAction(command_name="break action result")]
        field_descriptor.id = "age"

        internal_form = PicklableMock(name="internal_form_mock")
        internal_form.description.fields.items = PicklableMock(return_value=[("age", field_descriptor)])
        internal_form.field.field_validator.requirement.check = PicklableMock(return_value=True)
        field = PicklableMock()
        field.description = field_descriptor
        field.value = 61
        internal_form.fields = {field_descriptor: field, "age": field}
        internal_form.is_valid = PicklableMock(return_value=True)

        start_form_mock = PicklableMagicMock(name="start_form_mock")

        composite_form = PicklableMagicMock(name="form_mock")
        composite_form.forms = start_form_mock
        composite_form.set_valid = PicklableMock()

        context_forms = PicklableMagicMock(name="context_form_mock")
        context_forms.new = PicklableMock(return_value=PicklableMock(forms={"my form key": internal_form},
                                                                     is_endless=PicklableMock(return_value=False)))

        current_node_mock = PicklableMock(name="current_node_mock")
        current_node_mock.form_key = "start_node_mock"
        current_node_mock.available_nodes = False

        user = PicklableMock()
        user.forms = context_forms
        user.scenario_models = PicklableMagicMock()
        user.scenario_models.__getitem__ = PicklableMock(name="scenario_mock", return_value=current_node_mock)

        text_preprocessing_result = None

        scenario = TreeScenario(items, 1)

        result = scenario.run(text_preprocessing_result, user)

        self.assertFalse(scenario.actions[0].called)
        self.assertEqual(result[0].name, "break action result")



# coding: utf-8
import unittest

from smart_kit.names import message_names


class MessageNamesTest1(unittest.TestCase):
    def test_message_names(self):
        self.assertIsNotNone(message_names.ANSWER_TO_USER)
        self.assertIsNotNone(message_names.ERROR)
        self.assertIsNotNone(message_names.NOTHING_FOUND)
        self.assertIsNotNone(message_names.MESSAGE_TO_SKILL)
        self.assertIsNotNone(message_names.LOCAL_TIMEOUT)
        self.assertIsNotNone(message_names.RUN_APP)
        self.assertIsNotNone(message_names.CLOSE_APP)
        self.assertIsNotNone(message_names.SERVER_ACTION)



# coding: utf-8
from core.model.base_user import BaseUser
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.utils.utils import convert_version_to_list_of_int
from typing import List, Optional, Dict, Any

from core.model.factory import factory

from core.basic_models.requirement.basic_requirements import Requirement, ComparisonRequirement
from core.basic_models.operators.operators import Operator
from core.text_preprocessing.base import BaseTextPreprocessingResult


class BaseContainsRequirement(Requirement):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(BaseContainsRequirement, self).__init__(items)

    @property
    def descr_to_check_in(self):
        return NotImplementedError

    def get_field(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser):
        return NotImplementedError

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return self.get_field(text_preprocessing_result, user) in self.descr_to_check_in


class ChannelRequirement(BaseContainsRequirement):
    channels = List[str]

    # should_process_message compatible
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(ChannelRequirement, self).__init__(items, id)
        self.channels = items["channels"]

    @property
    def descr_to_check_in(self):
        return set(self.channels)

    def get_field(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser) -> str:
        return user.message.channel


class PlatformTypeRequirement(Requirement):

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(PlatformTypeRequirement, self).__init__(items, id)
        items = items or {}
        self.platfrom_type = items["platfrom_type"]

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return user.message.device.platform_type == self.platfrom_type


class BasicVersionRequirement(ComparisonRequirement):

    @factory(Operator)
    def build_operator(self):
        operator = dict(self._operator)
        operator["amount"] = convert_version_to_list_of_int(operator["amount"])
        return operator


class PlatformVersionRequirement(BasicVersionRequirement):

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        platform_version = convert_version_to_list_of_int(user.message.device.platform_version)
        return self.operator.compare(platform_version) if platform_version is not None else False


class SurfaceRequirement(Requirement):

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(SurfaceRequirement, self).__init__(items, id)
        items = items or {}
        self.surface = items["surface"]

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return user.message.device.surface == self.surface


class SurfaceVersionRequirement(BasicVersionRequirement):

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        surface_version = convert_version_to_list_of_int(user.message.device.surface_version)
        return self.operator.compare(surface_version) if surface_version is not None else False


class AppTypeRequirement(Requirement):

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(AppTypeRequirement, self).__init__(items, id)
        items = items or {}
        self.app_type = items["app_type"]

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return self.app_type in user.message.device.features.get("appTypes", [])


class CapabilitiesPropertyAvailableRequirement(Requirement):

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CapabilitiesPropertyAvailableRequirement, self).__init__(items, id)
        items = items or {}
        self.property_type = items["property_type"]

    def check(self, text_preprocessing_result: TextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        return user.message.device.capabilities.get(self.property_type, {}).get("available", False)



from typing import Dict

import nltk

from collections import defaultdict

from core.names.field import MESSAGE, ANNOTATIONS
from core.text_preprocessing.grammem.grammem_constants import TOKEN_VALUE, TOKEN_TYPE, IS_BEGINNING_OF_COMPOSITE, \
    COMPOSITE_TOKEN_VALUE, COMPOSITE_TOKEN_TYPE, VALUE, LIST_OF_TOKEN_TYPES_DATA
from core.text_preprocessing.helpers import TokenizeHelper
from core.text_preprocessing.constants import NUM_TOKEN, CCY_TOKEN, PERSON_TOKEN, MONEY_TOKEN, TIME_DATE_TOKEN, \
    PERIOD_TOKEN, TIME_DATE_INTERVAL_TOKEN, RELATIVE_TIME_TOKEN, ORG_TOKEN, GEO_TOKEN, TIME_TIME_TOKEN, \
    SENTENCE_ENDPOINT_TOKEN
from core.text_preprocessing.base import BaseTextPreprocessingResult


class TextPreprocessingResult(BaseTextPreprocessingResult):
    __slots__ = ("original_text", "normalized_text", "original_message_name", "__tokenized_elements_list",
                 "_tokenized_string", "_normalized_text_with_verb_mood", "_words_tokenized", "_words_tokenized_set",
                 "_num_token_values", "_person_token_values", "_money_token_values", "_time_date_token_values",
                 "_period_token_values", "_org_token_values", "_time_date_interval_token_values",
                 "_relative_token_values", "_ccy_token_values", "_geo_token_values", "pipeline_results",
                 "_tokenized_string_stop_words", "_words_tokenized_stop_words", "_words_tokenized_set_stop_words",
                 "model_ner", "dict_ner")

    @staticmethod
    def from_payload(payload: Dict):
        return TextPreprocessingResult(payload.get(MESSAGE, {}), payload.get(ANNOTATIONS, {}))

    def __init__(self, *args: Dict):
        items = {k: v for arg in args for k, v in arg.items()}
        super(TextPreprocessingResult, self).__init__(items)
        self.original_text = items.get("original_text", "")
        self.normalized_text = items.get("normalized_text", "")
        self.original_message_name = items.get("original_message_name", "")
        # TODO: погрумить с IR как объединить ner-ы
        self.model_ner = items.get("ner_prediction", [])
        self.dict_ner = items.get("entities", [])
        self.__tokenized_elements_list = items.get("tokenized_elements_list", [])
        self.pipeline_results = items.get("pipeline_results", {})
        self._human_normalized_text = items.get("human_normalized_text")
        self._human_normalized_text_with_anaphora = items.get("human_normalized_text_with_anaphora")
        self._tokenized_elements_list_pymorphy = None
        self._normalized_text_pymorphy = None
        self._tokenized_string = None
        self._normalized_text_with_verb_mood = None
        self._words_tokenized = None
        self._words_tokenized_set = None
        self._num_token_values = None
        self._person_token_values = None
        self._money_token_values = None
        self._time_date_token_values = None
        self._period_token_values = None
        self._org_token_values = None
        self._geo_token_values = None
        self._time_time_token_values = None
        self._time_date_interval_token_values = None
        self._relative_token_values = None
        self._ccy_token_values = None
        self._tokenized_string_stop_words = None
        self._words_tokenized_stop_words = None
        self._words_tokenized_set_stop_words = None

    @property
    def tokenized_elements_list_pymorphy(self):
        if self._tokenized_elements_list_pymorphy is None:

            from smart_kit.configs import get_app_config
            app_config = get_app_config()

            self._tokenized_elements_list_pymorphy = app_config.NORMALIZER(self.original_text)["tokenized_elements_list"]

        return self._tokenized_elements_list_pymorphy

    @property
    def normalized_text_pymorphy(self):
        if self._normalized_text_pymorphy is None:

            from smart_kit.configs import get_app_config
            app_config = get_app_config()

            normalized_words = [app_config.NORMALIZER.morph.pymorphy_analyzer.parse(tokenized_word)[0].normalized.normal_form
                                for tokenized_word in nltk.tokenize.word_tokenize(self.original_text)]
            self._normalized_text_pymorphy = " ".join(normalized_words)

        return self._normalized_text_pymorphy

    @property
    def tokenized_string(self):
        if self._tokenized_string is None:
            self._tokenized_string = TokenizeHelper.return_lemmas_only(self.__tokenized_elements_list,
                                                                       include_sentence_endpoint=False)
        return self._tokenized_string

    @property
    def tokenized_string_stop_words(self):
        if self._tokenized_string_stop_words is None:
            self._tokenized_string_stop_words = TokenizeHelper.return_lemmas_only(self.__tokenized_elements_list,
                                                                                  include_sentence_endpoint=False,
                                                                                  consider_stop_words=True)
        return self._tokenized_string_stop_words

    @property
    def human_normalized_text(self):
        if self._human_normalized_text is None:
            self._human_normalized_text = TokenizeHelper.get_human_normalized_text(self.__tokenized_elements_list)
        return self._human_normalized_text

    @property
    def human_normalized_text_with_anaphora(self):
        if self._human_normalized_text_with_anaphora is None:
            self._human_normalized_text_with_anaphora = \
                TokenizeHelper.get_human_normalized_text_with_anaphora(self.__tokenized_elements_list)
        return self._human_normalized_text_with_anaphora

    @property
    def normalized_text_with_verb_mood(self):
        if self._normalized_text_with_verb_mood is None:
            self._normalized_text_with_verb_mood = TokenizeHelper.return_lemmas_only(self.__tokenized_elements_list,
                                                                                     include_sentence_endpoint=True,
                                                                                     show_verb_mood=True)
        return self._normalized_text_with_verb_mood

    @property
    def words_tokenized(self):
        if self._words_tokenized is None:
            self._words_tokenized = self.tokenized_string.split(" ")
        return self._words_tokenized

    @property
    def words_tokenized_stop_words(self):
        if self._words_tokenized_stop_words is None:
            self._words_tokenized_stop_words = self.tokenized_string_stop_words.split(" ")
        return self._words_tokenized_stop_words

    @property
    def words_tokenized_set(self):
        if self._words_tokenized_set is None:
            self._words_tokenized_set = set(self.words_tokenized)
        return self._words_tokenized_set

    @property
    def words_tokenized_set_stop_words(self):
        if self._words_tokenized_set_stop_words is None:
            self._words_tokenized_set_stop_words = set(self.words_tokenized_stop_words)
        return self._words_tokenized_set_stop_words

    @property
    def all_token_values(self):
        result = defaultdict(list)
        for token in self.__tokenized_elements_list:
            for token_option in token.get(LIST_OF_TOKEN_TYPES_DATA, []):
                if token_option[TOKEN_TYPE] != SENTENCE_ENDPOINT_TOKEN:
                    result[token_option[TOKEN_TYPE]].append(token_option[TOKEN_VALUE])
            if token.get(IS_BEGINNING_OF_COMPOSITE):
                result[token.get(COMPOSITE_TOKEN_TYPE)].append(token.get(COMPOSITE_TOKEN_VALUE))
        return result

    def get_token_values_by_type(self, token_type):
        final_list = []
        for t in self.__tokenized_elements_list:
            list_of_values = t.get(LIST_OF_TOKEN_TYPES_DATA, [])
            for values_dicti in list_of_values:
                if values_dicti[TOKEN_TYPE] == token_type:
                    final_list.append(values_dicti.get(TOKEN_VALUE, {}).get(VALUE))
        return final_list

    def get_composite_token_values_by_type(self, composite_token_type):
        list_of_composite_token_values = []
        for token in self.__tokenized_elements_list:
            if token.get(COMPOSITE_TOKEN_TYPE) == composite_token_type and token.get(IS_BEGINNING_OF_COMPOSITE):
                list_of_composite_token_values.append(token.get(COMPOSITE_TOKEN_VALUE))
        return list_of_composite_token_values

    @property
    def num_token_values(self):
        if self._num_token_values is None:
            self._num_token_values = self.get_token_values_by_type(NUM_TOKEN)
        return self._num_token_values

    @property
    def person_token_values(self):
        if self._person_token_values is None:
            self._person_token_values = self.get_composite_token_values_by_type(PERSON_TOKEN)
        return self._person_token_values

    @property
    def money_token_values(self):
        if self._money_token_values is None:
            self._money_token_values = self.get_composite_token_values_by_type(MONEY_TOKEN)
        return self._money_token_values

    @property
    def time_date_token_values(self):
        if self._time_date_token_values is None:
            self._time_date_token_values = self.get_composite_token_values_by_type(TIME_DATE_TOKEN)
        return self._time_date_token_values

    @property
    def period_token_values(self):
        if self._period_token_values is None:
            self._period_token_values = self.get_composite_token_values_by_type(PERIOD_TOKEN)
        return self._period_token_values

    @property
    def org_token_values(self):
        if self._org_token_values is None:
            self._org_token_values = self.get_composite_token_values_by_type(ORG_TOKEN)
        return self._org_token_values

    @property
    def geo_token_values(self):
        if self._geo_token_values is None:
            self._geo_token_values = self.get_composite_token_values_by_type(GEO_TOKEN)
        return self._geo_token_values

    @property
    def time_date_interval_token_values(self):
        if self._time_date_interval_token_values is None:
            self._time_date_interval_token_values = self.get_composite_token_values_by_type(TIME_DATE_INTERVAL_TOKEN)
        return self._time_date_interval_token_values

    @property
    def time_time_token_values(self):
        if self._time_time_token_values is None:
            self._time_time_token_values = self.get_composite_token_values_by_type(TIME_TIME_TOKEN)
        return self._time_time_token_values

    @property
    def relative_token_values(self):
        if self._relative_token_values is None:
            self._relative_token_values = self.get_composite_token_values_by_type(RELATIVE_TIME_TOKEN)
        return self._relative_token_values

    @property
    def number_of_numbers(self):
        return len(self.num_token_values)

    @property
    def ccy_token_values(self):
        if self._ccy_token_values is None:
            self._ccy_token_values = self.get_token_values_by_type(CCY_TOKEN)
        return self._ccy_token_values

    @property
    def currencies_number(self):
        return len(self.ccy_token_values)

    @property
    def tokenized_elements_list(self):
        return self.__tokenized_elements_list

    @property
    def raw(self):
        return {"original_text": self.original_text,
                "normalized_text": self.normalized_text,
                "tokenized_elements_list": self.__tokenized_elements_list,
                "ner_prediction": self.model_ner,
                "entities": self.dict_ner}



from typing import List, Dict, Any, Optional

import scenarios.logging.logger_constants as log_const
from core.basic_models.actions.command import Command

from core.names.field import SERVER_ACTION
from core.logging.logger_utils import log
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.utils.pickle_copy import pickle_deepcopy
from scenarios.user.user_model import User

from smart_kit.handlers.handler_base import HandlerBase
from core.monitoring.monitoring import monitoring


class HandlerServerAction(HandlerBase):
    handler_name = "HandlerServerAction"

    def __init__(self, app_name: str, action_name: Optional[str] = None):
        super().__init__(app_name)
        self._action_name = action_name

    def get_action_name(self, payload: Dict[str, Any], user: User):
        return payload[SERVER_ACTION]["action_id"]

    def get_action_params(self, payload: Dict[str, Any]):
        return payload[SERVER_ACTION].get("parameters", {})

    def run(self, payload: Dict[str, Any], user: User) -> List[Command]:
        commands = super().run(payload, user)
        action_params = pickle_deepcopy(self.get_action_params(payload))
        params = {log_const.KEY_NAME: "handling_server_action",
                  "server_action_params": str(action_params),
                  "server_action_id": self.get_action_name(payload, user)}
        log("HandlerServerAction %(server_action_id)s started", user, params)

        app_info = user.message.app_info
        monitoring.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                    user, app_info=app_info)

        action_id = self.get_action_name(payload, user)
        action = user.descriptions["external_actions"][action_id]
        commands.extend(action.run(user, TextPreprocessingResult({}), action_params))
        return commands



import sys
from functools import wraps


def exc_handler(on_error_obj_method_name=None, handled_exceptions=None):
    handled_exceptions = tuple(handled_exceptions) if handled_exceptions else (Exception,)

    def exc_handler_decorator(funct):
        @wraps(funct)
        def _wrapper(obj, *args, **kwarg):
            result = None
            try:
                result = funct(obj, *args, **kwarg)
            except handled_exceptions:
                try:
                    on_error = getattr(obj, on_error_obj_method_name) if \
                        on_error_obj_method_name else (lambda *x: None)
                    result = on_error(*args, **kwarg)
                except:
                    print(sys.exc_info())
            return result

        return _wrapper

    return exc_handler_decorator



# coding: utf-8
import unittest
from unittest.mock import Mock

from core.basic_models.requirement import device_requirements


class RequirementTest1(unittest.TestCase):
    def setUp(self):
        self.test_items1 = {"platfrom_type": "any platform"}  # PLATFROM - так задумано?
        self.test_items2 = {"platform_type": "any platform 2"}
        self.test_id = "12345"
        self.test_text_processing_result = Mock('Text processing result')
        self.test_user1 = Mock('User')
        self.test_user1.message = Mock('Message')
        self.test_user1.message.device = Mock('Device')
        self.test_user1.message.device.platform_type = "any platform"
        self.test_user2 = Mock('User')
        self.test_user2.message = Mock('Message')
        self.test_user2.message.device = Mock('Device')
        self.test_user2.message.device.platform_type = "any platform 2"

    def test_platform_type_requirement_init(self):
        obj1 = device_requirements.PlatformTypeRequirement(self.test_items1, self.test_id)
        self.assertTrue(obj1.platfrom_type == self.test_items1["platfrom_type"])
        self.assertTrue(obj1.id == self.test_id)
        with self.assertRaises(KeyError):
            obj2 = device_requirements.PlatformTypeRequirement("", self.test_id)
            self.assertTrue(obj2.items == {})
        with self.assertRaises(KeyError):
            obj3 = device_requirements.PlatformTypeRequirement(self.test_items2, self.test_id)

    def test_platform_type_requirement_check(self):
        obj1 = device_requirements.PlatformTypeRequirement(self.test_items1, self.test_id)
        self.assertTrue(obj1.check(self.test_text_processing_result, self.test_user1))
        self.assertTrue(not obj1.check(self.test_text_processing_result, self.test_user2))


class RequirementTest2(unittest.TestCase):
    def setUp(self):
        self.test_items1 = {"operator": {"amount": "1.11.111"}}
        self.test_items2 = {"operator": {"amount": "AAA.11.111"}}
        self.test_id1 = "12345"

    def test_basic_version_requirement_init(self):
        obj1 = device_requirements.BasicVersionRequirement(self.test_items1, self.test_id1)
        self.assertTrue(obj1._operator == self.test_items1["operator"])
        self.assertTrue(obj1.id == self.test_id1)




from unittest import TestCase

from core.basic_models.actions.command import Command


class TestCommand(TestCase):
    def test_1(self):
        expected = {"messageName": "my_name", "payload": {}}

        command = Command("my_name")
        result = command.raw

        self.assertDictEqual(expected, result)

    def test_2(self):
        expected = {"messageName": "my_name", "payload": {"id": 5}}

        command = Command("my_name", {"id": 5})
        result = command.raw

        self.assertDictEqual(expected, result)



"""Requirements на текст запроса пользователя."""

from typing import Optional, Dict, Any

from core.basic_models.requirement.basic_requirements import Requirement, ComparisonRequirement
from core.logging.logger_utils import log
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult
from scenarios.user.user_model import User


class AnySubstringInLoweredTextRequirement(Requirement):
    """Условие возвращает True, если хотя бы одна подстрока из списка substrings встречается
    в оригинальном тексте в нижнем регистре, иначе - False.
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(AnySubstringInLoweredTextRequirement, self).__init__(items, id)
        self.substrings = self.items["substrings"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        lowered_text = text_preprocessing_result.lower() if isinstance(text_preprocessing_result, str) \
            else text_preprocessing_result.raw["original_text"].lower()
        return any(s.lower() in lowered_text for s in self.substrings)


class NormalizedInputWordsRequirement(Requirement):

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(NormalizedInputWordsRequirement, self).__init__(items, id)

        # Получаем используемый нормализатор из конфига аппа
        from smart_kit.configs import get_app_config
        app_config = get_app_config()
        self.normalizer = app_config.NORMALIZER

        # Нормализуем входные слова из условия
        self.input_words = set(items["input_words"])
        self.normalized_input_words = set([
            norm_res["normalized_text"].replace(".", "").strip()
            for norm_res in self.normalizer.normalize_sequence(list(self.input_words))
        ])


class IntersectionWithTokensSetRequirement(NormalizedInputWordsRequirement):
    """Условие возвращает True, если хотя бы одно слово из нормализованного вида запроса входит
    в список слов input_words, иначе - False.
    Слова из input_words также проходят нормализацию перед сравнением.
    """

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        words_normalized_set = set([
            token["lemma"] for token in text_preprocessing_result.raw["tokenized_elements_list_pymorphy"]
            if not token.get("token_type") == "SENTENCE_ENDPOINT_TOKEN"
        ])
        result = bool(self.normalized_input_words.intersection(words_normalized_set))
        if result:
            params = self._log_params()
            params["normalized_input_words"] = self.normalized_input_words
            params["words_normalized_set"] = words_normalized_set
            message = "Requirement: %(requirement)s, normalized_input_words: %(normalized_input_words)s, " \
                      "words_normalized_set: %(words_normalized_set)s"
            log(message, user, params)
        return result


class NormalizedTextInSetRequirement(NormalizedInputWordsRequirement):
    """Условие возвращает True, если в нормализованном представлении запрос полностью совпадает с одной из
    нормализованных строк из input_words, иначе - False.
    """

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        normalized_text = text_preprocessing_result.raw["normalized_text"].replace(".", "").strip()
        result = normalized_text in self.normalized_input_words
        if result:
            params = self._log_params()
            params["normalized_text"] = normalized_text
            params["normalized_input_words"] = self.normalized_input_words
            message = "Requirement: %(requirement)s, normalized_text: %(normalized_text)s, " \
                      "normalized_input_words: %(normalized_input_words)s"
            log(message, user, params)

        return result


class PhoneNumberNumberRequirement(ComparisonRequirement):
    """Условие возвращает True, если кол-во номеров телефонов больше/меньше/.. X, иначе - False.
    Строго говоря, считается кол-во токенов, имеющих token_type = "PHONE_NUMBER_TOKEN".
    """

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super().__init__(items, id)

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        len_phone_number_token = len(text_preprocessing_result.get_token_values_by_type("PHONE_NUMBER_TOKEN"))
        result = self.operator.compare(len_phone_number_token)
        if result:
            params = self._log_params()
            params["len_phone_number_token"] = len_phone_number_token
            message = "Requirement: %(requirement)s, len_phone_number_token: %(len_phone_number_token)s"
            log(message, user, params)
        return result


class NumInRangeRequirement(Requirement):
    """Условие возвращает True, если число находится в заданном диапазоне, иначе - False."""

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(NumInRangeRequirement, self).__init__(items, id)
        self.min_num = float(items["min_num"])
        self.max_num = float(items["max_num"])

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: User,
              params: Dict[str, Any] = None) -> bool:
        num = float(text_preprocessing_result.num_token_values)
        return self.min_num <= num <= self.max_num if num else False



import collections
import json
from typing import Optional, Dict, Any, Union, List

from jinja2 import exceptions as jexcept

from core.basic_models.actions.basic_actions import Action
from core.basic_models.actions.command import Command
from core.basic_models.parametrizers.parametrizer import BasicParametrizer
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.unified_template.unified_template import UnifiedTemplate


class BaseSetVariableAction(Action):
    key: str
    loader: Optional[str]
    loaders = collections.defaultdict(str, {"json": json.loads, "float": float, "int": int})
    value: Union[str, Dict]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.key: str = items["key"]
        self.loader = items.get('loader')
        value: str = items["value"]
        self.template: UnifiedTemplate = UnifiedTemplate(value)

    def _set(self, user, value):
        raise NotImplemented

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        params = user.parametrizer.collect(text_preprocessing_result)
        try:
            # if path is wrong, it may fail with UndefinedError
            # notion: {key: None} will return "None";
            # not existing key or value "" will return ""; otherwise question in scenario will go in cycles
            value = self.template.render(params)
        except jexcept.UndefinedError:
            value = None

        if self.loader:
            if value:
                loader = self.loaders[self.loader]
                value = loader(value)
            else:
                value = None

        self._set(user, value)
        return commands


class SetVariableAction(BaseSetVariableAction):
    version: Optional[int]
    parametrizer: BasicParametrizer
    loader: Optional[str]
    key: str
    loaders = collections.defaultdict(str, {"json": json.loads, "float": float, "int": int})
    ttl: int
    value: Union[str, Dict]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.ttl: int = items.get("ttl")

    def _set(self, user, value):
        user.variables.set(self.key, value, self.ttl)


class DeleteVariableAction(Action):
    version: Optional[int]
    key: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.key: str = items["key"]

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.variables.delete(self.key)
        return commands


class ClearVariablesAction(Action):
    version: Optional[int]

    def __init__(self, items: Dict[str, Any] = None, id: Optional[str] = None):
        super().__init__(items, id)

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.variables.clear()
        return commands


class SetLocalVariableAction(BaseSetVariableAction):
    def _set(self, user, value):
        user.local_vars.set(self.key, value)



# coding: utf-8
import os

import yaml

from core.configs.base_config import BaseConfig
from core.repositories.file_repository import FileRepository
from smart_kit.utils.logger_writer.logger_formatter import SmartKitJsonFormatter

SmartKitJsonFormatter.VERSION = os.getenv("VERSION") or 0

LOGGING_CONFIG = "logging_config"


class LoggerConfig(BaseConfig):
    def __init__(self, config_path):
        self.config_path = config_path
        super(LoggerConfig, self).__init__()
        self.repositories = [
            FileRepository(self.subfolder_path("logging_config.yml"),
                           loader=yaml.safe_load, key=LOGGING_CONFIG)
        ]

    @property
    def _subfolder(self):
        return self.config_path



import os
import unittest
from time import time
from unittest.mock import Mock, patch

import smart_kit
from core.basic_models.classifiers.basic_classifiers import ExternalClassifier
from core.basic_models.operators.operators import Operator
from core.basic_models.requirement.basic_requirements import Requirement, CompositeRequirement, AndRequirement, \
    OrRequirement, NotRequirement, RandomRequirement, TopicRequirement, TemplateRequirement, RollingRequirement, \
    TimeRequirement, DateTimeRequirement, IntersectionRequirement, ClassifierRequirement, FormFieldValueRequirement, \
    EnvironmentRequirement, CharacterIdRequirement, FeatureToggleRequirement
from core.basic_models.requirement.counter_requirements import CounterValueRequirement, CounterUpdateTimeRequirement
from core.basic_models.requirement.device_requirements import ChannelRequirement
from core.basic_models.requirement.user_text_requirements import AnySubstringInLoweredTextRequirement, \
    PhoneNumberNumberRequirement, NumInRangeRequirement, IntersectionWithTokensSetRequirement, \
    NormalizedTextInSetRequirement
from core.model.registered import registered_factories
from smart_kit.text_preprocessing.local_text_normalizer import LocalTextNormalizer
from smart_kit.utils.picklable_mock import PicklableMock


def patch_get_app_config(mock_get_app_config):
    result = PicklableMock()
    sk_path = os.path.dirname(smart_kit.__file__)
    result.STATIC_PATH = os.path.join(sk_path, 'template/static')
    mock_get_app_config.return_value = result
    result.NORMALIZER = LocalTextNormalizer()
    result.ENVIRONMENT = "ift"
    mock_get_app_config.return_value = result


class MockRequirement:
    def __init__(self, items=None):
        items = items or {}
        self.cond = items.get("cond") or False

    def check(self, text_preprocessing_result, user, params):
        return self.cond


class MockTextNormalizationResult:
    def __init__(self, normalized=None, number_of_numbers=None, currencies_number=None, tokens=None):
        number_of_numbers = number_of_numbers or 0
        currencies_number = currencies_number or 0
        if normalized:
            self.words_tokenized = list(token.get("text") for token in normalized)
            self.words_tokenized_set = set(self.words_tokenized)
        else:
            self.words_tokenized = list()
            self.words_tokenized_set = set()
        self.number_of_numbers = number_of_numbers
        self.currencies_number = currencies_number
        self.tokenized_elements_list = tokens


class MockAmountOperator:
    def __init__(self, items):
        self.amount = items["amount"]

    def compare(self, value):
        return value > self.amount


class MockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value > self.amount


class EQMockOperator:
    def __init__(self, amount):
        self.amount = amount

    def compare(self, value):
        return value == self.amount


class RequirementTest(unittest.TestCase):
    def test_base(self):
        requirement = Requirement(None)
        assert requirement.check(None, None)

    def test_composite(self):
        registered_factories[Requirement] = MockRequirement
        requirement = CompositeRequirement({"requirements": [
            {"cond": True},
            {"cond": True}
        ]})
        self.assertEqual(len(requirement.requirements), 2)
        self.assertTrue(requirement.check(None, None))

    def test_and_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = AndRequirement({"requirements": [
            {"cond": True},
            {"cond": True}
        ]})
        self.assertTrue(requirement.check(None, None))

    def test_and_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = AndRequirement({"requirements": [
            {"cond": True},
            {"cond": False}
        ]})
        self.assertFalse(requirement.check(None, None))

    def test_or_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = OrRequirement({"requirements": [
            {"cond": True},
            {"cond": False}
        ]})
        self.assertTrue(requirement.check(None, None))

    def test_or_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = OrRequirement({"requirements": [
            {"cond": False},
            {"cond": False}
        ]})
        self.assertFalse(requirement.check(None, None))

    def test_not_success(self):
        registered_factories[Requirement] = MockRequirement
        requirement = NotRequirement({"requirement": {"cond": False}})
        self.assertTrue(requirement.check(None, None))

    def test_not_fail(self):
        registered_factories[Requirement] = MockRequirement
        requirement = NotRequirement({"requirement": {"cond": True}})
        self.assertFalse(requirement.check(None, None))

    def test_channel_success(self):
        user = PicklableMock()
        message = Mock(channel="ch1")
        user.message = message
        requirement = ChannelRequirement({"channels": ["ch1"]})
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_channel_fail(self):
        user = PicklableMock()
        message = Mock(channel="ch2")
        user.message = message
        requirement = ChannelRequirement({"channels": ["ch1"]})
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_random_requirement_true(self):
        requirement = RandomRequirement({"percent": 100})
        self.assertTrue(requirement.check(None, None))

    def test_random_requirement_false(self):
        requirement = RandomRequirement({"percent": 0})
        self.assertFalse(requirement.check(None, None))

    def test_topic_requirement(self):
        requirement = TopicRequirement({"topics": ["test"]})
        user = PicklableMock()
        message = PicklableMock()
        message.topic_key = "test"
        user.message = message
        self.assertTrue(requirement.check(None, user))

    def test_counter_value_requirement(self):
        registered_factories[Operator] = MockAmountOperator
        user = PicklableMock()
        counter = PicklableMock()
        counter.__gt__ = Mock(return_value=True)
        user.counters = {"test": counter}
        requirement = CounterValueRequirement({"operator": {"type": "equal", "amount": 2}, "key": "test"})
        self.assertTrue(requirement.check(None, user))

    def test_counter_time_requirement(self):
        registered_factories[Operator] = MockAmountOperator
        user = PicklableMock()
        counter = PicklableMock()
        counter.update_time = int(time()) - 10
        user.counters = {"test": counter}
        requirement = CounterUpdateTimeRequirement({"operator": {"type": "more_or_equal", "amount": 5}, "key": "test"})
        self.assertTrue(requirement.check(None, user))

    def test_template_req_true(self):
        items = {
            "template": "{{ payload.message.strip() in payload.murexIds }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {
            "groupCode": "BROKER",
            "murexIds": ["AAA", "BBB"],
            "message": " BBB    "
        }}
        user = PicklableMock()
        user.parametrizer = PicklableMock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertTrue(requirement.check(None, user))

    def test_template_req_false(self):
        items = {
            "template": "{{ payload.groupCode == 'BROKER' }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {"groupCode": "BROKER1"}}
        user = PicklableMock()
        user.parametrizer = PicklableMock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertFalse(requirement.check(None, user))

    def test_template_req_raise(self):
        items = {
            "template": "{{ payload.groupCode }}"
        }
        requirement = TemplateRequirement(items)
        params = {"payload": {"groupCode": "BROKER1"}}
        user = PicklableMock()
        user.parametrizer = PicklableMock()
        user.parametrizer.collect = Mock(return_value=params)
        self.assertRaises(TypeError, requirement.check, None, user)

    def test_rolling_requirement_true(self):
        user = PicklableMock()
        user.id = "353454"
        requirement = RollingRequirement({"percent": 100})
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_rolling_requirement_false(self):
        user = PicklableMock()
        user.id = "353454"
        requirement = RollingRequirement({"percent": 0})
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_time_requirement_true(self):
        user = PicklableMock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610990255000,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = TimeRequirement(
            {
                "operator": {
                    "type": "more",
                    "amount": "17:00:00",
                }
            }
        )
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_time_requirement_false(self):
        user = PicklableMock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = TimeRequirement(
            {
                "operator": {
                    "type": "more",
                    "amount": "18:00:00",
                }
            }
        )
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    def test_datetime_requirement_true(self):
        user = PicklableMock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = DateTimeRequirement(
            {
                "match_cron": "*/17 14-19 * * mon"
            }
        )
        text_normalization_result = None
        self.assertTrue(requirement.check(text_normalization_result, user))

    def test_datetime_requirement_false(self):
        user = PicklableMock()
        user.id = "353454"
        user.message.payload = {
            "meta": {
                "time": {
                    "timestamp": 1610979455663,  # ~ 2021-01-18 17:17:35
                    "timezone_offset_sec": 1000000000,  # shouldn't affect
                }
            }
        }
        requirement = DateTimeRequirement(
            {
                "match_cron": "* * * * 6,7"
            }
        )
        text_normalization_result = None
        self.assertFalse(requirement.check(text_normalization_result, user))

    @patch('smart_kit.configs.get_app_config')
    def test_intersection_requirement_true(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        user = PicklableMock()
        requirement = IntersectionRequirement(
            {
                "phrases": [
                    'да',
                    'давай',
                    'хочу',
                ]
            }
        )
        text_normalization_result = PicklableMock()
        text_normalization_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'я'},
            {'lemma': 'хотеть'},
        ]
        self.assertTrue(requirement.check(text_normalization_result, user))

    @patch('smart_kit.configs.get_app_config')
    def test_intersection_requirement_false(self, mock_get_app_config):
        patch_get_app_config(mock_get_app_config)
        user = PicklableMock()
        requirement = IntersectionRequirement(
            {
                "phrases": [
                    'да',
                    'давай',
                    'хочу',
                ]
            }
        )
        text_normalization_result = PicklableMock()
        text_normalization_result.tokenized_elements_list_pymorphy = [
            {'lemma': 'ни'},
            {'lemma': 'за'},
            {'lemma': 'что'},
        ]
        self.assertFalse(requirement.check(text_normalization_result, user))

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[{"answer": "нет", "score": 1.0, "other": False}])
    def test_classifier_requirement_true(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает True, если результат классификации запроса относится к одной
        из указанных категорий, прошедших порог, но не равной классу other.
        """
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = PicklableMock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(PicklableMock(), mock_user)
        self.assertTrue(result)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[])
    def test_classifier_requirement_false(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает False, если модель классификации не вернула ответ."""
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = PicklableMock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(PicklableMock(), mock_user)
        self.assertFalse(result)

    @patch.object(ExternalClassifier, "find_best_answer", return_value=[{"answer": "other", "score": 1.0, "other": True}])
    def test_classifier_requirement_false_if_class_other(self, mock_classifier_model):
        """Тест кейз проверяет что условие возвращает False, если наиболее вероятный вариант есть класс other."""
        test_items = {"type": "classifier", "classifier": {"type": "external", "classifier": "hello_scenario_classifier"}}
        classifier_requirement = ClassifierRequirement(test_items)
        mock_user = PicklableMock()
        mock_user.descriptions = {"external_classifiers": ["read_book_or_not_classifier", "hello_scenario_classifier"]}
        result = classifier_requirement.check(PicklableMock(), mock_user)
        self.assertFalse(result)

    def test_form_field_value_requirement_true(self):
        """Тест кейз проверяет что условие возвращает True, т.к в
        форме form_name в поле form_field значение совпадает с переданным field_value.
        """
        form_name = "test_form"
        form_field = "test_field"
        field_value = "test_value"

        test_itmes = {"form_name": form_name, "field_name": form_field, "value": field_value}
        req_form_field_value = FormFieldValueRequirement(test_itmes)

        user = PicklableMock()
        user.forms = {form_name: PicklableMock()}
        user.forms[form_name].fields = {form_field: PicklableMock(), "value": field_value}
        user.forms[form_name].fields[form_field].value = field_value

        result = req_form_field_value.check(PicklableMock(), user)
        self.assertTrue(result)

    def test_form_field_value_requirement_false(self):
        """Тест кейз проверяет что условие возвращает False, т.к в
        форме form_name в поле form_field значение НЕ совпадает с переданным field_value.
        """
        form_name = "test_form"
        form_field = "test_field"
        field_value = "test_value"

        test_itmes = {"form_name": form_name, "field_name": form_field, "value": field_value}
        req_form_field_value = FormFieldValueRequirement(test_itmes)

        user = PicklableMock()
        user.forms = {form_name: PicklableMock()}
        user.forms[form_name].fields = {form_field: PicklableMock(), "value": "OTHER_TEST_VAL"}
        user.forms[form_name].fields[form_field].value = "OTHER_TEST_VAL"

        result = req_form_field_value.check(PicklableMock(), user)
        self.assertFalse(result)

    @patch("smart_kit.configs.get_app_config")
    def test_environment_requirement_true(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает True, т.к среда исполнения из числа values."""
        patch_get_app_config(mock_get_app_config)
        environment_req = EnvironmentRequirement({"values": ["ift", "uat"]})
        self.assertTrue(environment_req.check(PicklableMock(), PicklableMock()))

    @patch("smart_kit.configs.get_app_config")
    def test_environment_requirement_false(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает False, т.к среда исполнения НЕ из числа values."""
        patch_get_app_config(mock_get_app_config)
        environment_req = EnvironmentRequirement({"values": ["uat", "pt"]})
        self.assertFalse(environment_req.check(PicklableMock(), PicklableMock()))

    def test_any_substring_in_lowered_text_requirement_true(self):
        """Тест кейз проверяет что условие возвращает True, т.к нашлась подстрока из списка substrings, которая
        встречается в оригинальном тексте в нижнем регистре.
        """
        req = AnySubstringInLoweredTextRequirement({"substrings": ["искомая подстрока", "другое знанчение"]})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"original_text": "КАКОЙ-ТО ТЕКСТ С ИСКОМАЯ ПОДСТРОКА"}
        result = req.check(text_preprocessing_result, PicklableMock())
        self.assertTrue(result)

    def test_any_substring_in_lowered_text_requirement_false(self):
        """Тест кейз проверяет что условие возвращает False, т.к НЕ нашлась ни одна подстрока из списка substrings,
        которая бы встречалась в оригинальном тексте в нижнем регистре.
        """
        req = AnySubstringInLoweredTextRequirement({"substrings": ["искомая подстрока", "другая подстрока"]})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"original_text": "КАКОЙ-ТО ТЕКСТ"}
        result = req.check(text_preprocessing_result, PicklableMock())
        self.assertFalse(result)

    def test_num_in_range_requirement_true(self):
        """Тест кейз проверяет что условие возвращает True, т.к число находится в заданном диапазоне."""
        req = NumInRangeRequirement({"min_num": "5", "max_num": "10"})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.num_token_values = 7
        self.assertTrue(req.check(text_preprocessing_result, PicklableMock()))

    def test_num_in_range_requirement_false(self):
        """Тест кейз проверяет что условие возвращает False, т.к число НЕ находится в заданном диапазоне."""
        req = NumInRangeRequirement({"min_num": "5", "max_num": "10"})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.num_token_values = 20
        self.assertFalse(req.check(text_preprocessing_result, PicklableMock()))

    def test_phone_number_number_requirement_true(self):
        """Тест кейз проверяет что условие возвращает True, т.к кол-во номеров телефонов больше заданного."""
        req = PhoneNumberNumberRequirement({"operator": {"type": "more", "amount": 1}})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.get_token_values_by_type.return_value = ["89030478799", "89092534523"]
        self.assertTrue(req.check(text_preprocessing_result, PicklableMock()))

    def test_phone_number_number_requirement_false(self):
        """Тест кейз проверяет что условие возвращает False, т.к кол-во номеров телефонов НЕ больше заданного."""
        req = PhoneNumberNumberRequirement({"operator": {"type": "more", "amount": 10}})
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.get_token_values_by_type.return_value = ["89030478799"]
        self.assertFalse(req.check(text_preprocessing_result, PicklableMock()))

    @patch("smart_kit.configs.get_app_config")
    def test_intersection_with_tokens_requirement_true(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает True, т.к хотя бы одно слово из нормализованного
        вида запроса входит в список слов input_words.
        """
        patch_get_app_config(mock_get_app_config)

        req = IntersectionWithTokensSetRequirement({"input_words": ["погода", "время"]})

        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"tokenized_elements_list_pymorphy": [
            {"text": "прогноз", "grammem_info": {
                "animacy": "inan", "case": "acc", "gender": "masc", "number": "sing", "raw_gram_info":
                    "animacy=inan|case=acc|gender=masc|number=sing", "part_of_speech": "NOUN"}, "lemma": "прогноз"},
            {"text": "погоды", "grammem_info": {
                "animacy": "inan", "case": "gen", "gender": "fem", "number": "sing",
                "raw_gram_info": "animacy=inan|case=gen|gender=fem|number=sing",
                "part_of_speech": "NOUN"}, "lemma": "погода"}
            ]}

        self.assertTrue(req.check(text_preprocessing_result, PicklableMock()))

    @patch("smart_kit.configs.get_app_config")
    def test_intersection_with_tokens_requirement_false(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает False, т.к ни одно слово из нормализованного
        вида запроса не входит в список слов input_words.
        """
        patch_get_app_config(mock_get_app_config)

        req = IntersectionWithTokensSetRequirement({"input_words": ["время"]})

        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"tokenized_elements_list_pymorphy": [
            {"text": "прогноз", "grammem_info": {
                "animacy": "inan", "case": "acc", "gender": "masc", "number": "sing", "raw_gram_info":
                    "animacy=inan|case=acc|gender=masc|number=sing", "part_of_speech": "NOUN"}, "lemma": "прогноз"},
            {"text": "погоды", "grammem_info": {
                "animacy": "inan", "case": "gen", "gender": "fem", "number": "sing",
                "raw_gram_info": "animacy=inan|case=gen|gender=fem|number=sing",
                "part_of_speech": "NOUN"}, "lemma": "погода"}
        ]}

        self.assertFalse(req.check(text_preprocessing_result, PicklableMock()))

    @patch("smart_kit.configs.get_app_config")
    def test_normalized_text_in_set_requirement_true(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает True, т.к в нормализованном представлении
        запрос полностью совпадает с одной из нормализованных строк из input_words.
        """
        patch_get_app_config(mock_get_app_config)

        req = NormalizedTextInSetRequirement({"input_words": ["погода", "время"]})

        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"normalized_text": "погода ."}

        self.assertTrue(req.check(text_preprocessing_result, PicklableMock()))

    @patch("smart_kit.configs.get_app_config")
    def test_normalized_text_in_set_requirement_false(self, mock_get_app_config):
        """Тест кейз проверяет что условие возвращает False, т.к в нормализованном представлении
        запрос НЕ совпадает ни с одной из нормализованных строк из input_words.
        """
        patch_get_app_config(mock_get_app_config)

        req = NormalizedTextInSetRequirement({"input_words": ["погода", "время"]})

        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.raw = {"normalized_text": "хотеть узнать ."}

        self.assertFalse(req.check(text_preprocessing_result, PicklableMock()))

    def test_character_id_requirement_true(self):
        req = CharacterIdRequirement({"values": ["sber", "afina"]})
        user = Mock()
        user.message = Mock()
        user.message.payload = {"character": {"id": "sber", "name": "Сбер", "gender": "male"}}
        self.assertTrue(req.check(Mock(), user))

    def test_character_id_requirement_false(self):
        req = CharacterIdRequirement({"values": ["afina"]})
        user = Mock()
        user.message = Mock()
        user.message.payload = {"character": {"id": "sber", "name": "Сбер", "gender": "male"}}
        self.assertFalse(req.check(Mock(), user))

    def test_feature_toggle_check_requirement_true(self):
        req = FeatureToggleRequirement({"toggle_name": "test_true_toggle_name"})
        mock_user = Mock()
        mock_user.settings = {"template_settings": {"test_true_toggle_name": True}}
        self.assertTrue(req.check(Mock(), mock_user))

    def test_feature_toggle_check_requirement_false(self):
        req = FeatureToggleRequirement({"toggle_name": "test_false_toggle_name"})
        mock_user = Mock()
        mock_user.settings = {"template_settings": {"test_false_toggle_name": False}}
        self.assertFalse(req.check(Mock(), mock_user))



# coding: utf-8

from scenarios.scenario_models.forms.form import form_model_factory
from core.model.lazy_items import LazyItems


class Forms(LazyItems):
    def __init__(self, items, descriptions, user):
        super(Forms, self).__init__(items, descriptions, user, form_model_factory)

    def __getitem__(self, description):
        if hasattr(description, "id"):
            id = description.id
        else:
            id = description
            description = self._descriptions.get(id)
        form = self._items.get(description)
        raw_data = self._raw_items.get(id)
        if form is None and raw_data is not None:
            form = self._build_factory(description, raw_data)
            self._items[description] = form
        return form

    def get_or_create(self, key):
        form = self[key]
        return form or self.new(key)

    def remove_item(self, descr_id):
        if descr_id in self._raw_items:
            self._raw_items.pop(descr_id)
            description = self._descriptions.get(descr_id)
            if description and description in self._items:
                self._items.pop(description)

    def new(self, descr_id):
        self.remove_item(descr_id)
        description = self._descriptions[descr_id]
        form = self._build_factory(description, {})
        self._items[description] = form
        form.touch()
        return form

    def collect_form_fields(self):
        fields = {}
        for form in self.raw:
            if form in self.descriptions:
                inner_form = self[form]
                data = inner_form.get_fields_values()
                fields.update(data)
        return fields

    def expire(self):
        to_remove = []
        descr_keys = self.descriptions.keys()
        for description_id in self.raw:
            if description_id in descr_keys:
                
                item = self[description_id]
                if item and item.check_expired():
                    to_remove.append(description_id)
            else:
                to_remove.append(description_id)
        for descr_id in to_remove:
            self.remove_item(descr_id)

    def clear_all(self):
        self._raw_items.clear()
        self._items.clear()

    def clear_form(self, scenario_name):
        scenario_descriptions = self._user.descriptions["scenarios"]
        if scenario_name in scenario_descriptions:
            scenario = scenario_descriptions[scenario_name]
            if not scenario.keep_forms_alive:
                self._user.forms.remove_item(scenario.form_type)



import re
from core.unified_template.num2text import Num2Text


class Num2Ordinal:
    orders = {3: {"m": {"nomn": "тысячный", "gent": "тысячного"},
                  "f": {"nomn": "тысячная", "gent": "тысячной"},
                  "n": {"nomn": "тысячное", "gent": "тысячного"}},
              6: {"m": {"nomn": "миллионный", "gent": "миллионного"},
                  "f": {"nomn": "миллионная", "gent": "миллионной"},
                  "n": {"nomn": "миллионное", "gent": "миллионного"}},
              9: {"m": {"nomn": "миллиардный", "gent": "миллиардного"},
                  "f": {"nomn": "миллиардная", "gent": "миллиардной"},
                  "n": {"nomn": "миллиардное", "gent": "миллиардного"}},
              12: {"m": {"nomn": "триллионный", "gent": "триллионного"},
                   "f": {"nomn": "триллионная", "gent": "триллионной"},
                   "n": {"nomn": "триллионное", "gent": "триллионного"}},
              15: {"m": {"nomn": "квадриллионный", "gent": "квадриллионного"},
                   "f": {"nomn": "квадриллионная", "gent": "квадриллионной"},
                   "n": {"nomn": "квадриллионное", "gent": "квадриллионного"}},
              18: {"m": {"nomn": "квадриллионный", "gent": "квадриллионного"},
                   "f": {"nomn": "квадриллионная", "gent": "квадриллионной"},
                   "n": {"nomn": "квадриллионное", "gent": "квадриллионного"}},
              21: {"m": {"nomn": "секстиллионный", "gent": "секстиллионного"},
                   "f": {"nomn": "секстиллионная", "gent": "секстиллионной"},
                   "n": {"nomn": "секстиллионное", "gent": "секстиллионного"}}}
    before_orders = {1: {"m": {"nomn": "одно", "gent": "одно"},
                         "f": {"nomn": "одно", "gent": "одно"},
                         "n": {"nomn": "одно", "gent": "одно"}},
                     2: {"m": {"nomn": "двух", "gent": "двух"},
                         "f": {"nomn": "двух", "gent": "двух"},
                         "n": {"nomn": "двух", "gent": "двух"}},
                     3: {"m": {"nomn": "трёх", "gent": "трёх"},
                         "f": {"nomn": "трёх", "gent": "трёх"},
                         "n": {"nomn": "трёх", "gent": "трёх"}},
                     4: {"m": {"nomn": "четырёх", "gent": "четырёх"},
                         "f": {"nomn": "четырёх", "gent": "четырёх"},
                         "n": {"nomn": "четырёх", "gent": "четырёх"}},
                     5: {"m": {"nomn": "пяти", "gent": "пяти"},
                         "f": {"nomn": "пяти", "gent": "пяти"},
                         "n": {"nomn": "пяти", "gent": "пяти"}},
                     6: {"m": {"nomn": "шести", "gent": "шести"},
                         "f": {"nomn": "шести", "gent": "шести"},
                         "n": {"nomn": "шести", "gent": "шести"}},
                     7: {"m": {"nomn": "семи", "gent": "семи"},
                         "f": {"nomn": "семи", "gent": "семи"},
                         "n": {"nomn": "семи", "gent": "семи"}},
                     8: {"m": {"nomn": "восьми", "gent": "восьми"},
                         "f": {"nomn": "восьми", "gent": "восьми"},
                         "n": {"nomn": "восьми", "gent": "восьми"}},
                     9: {"m": {"nomn": "девяти", "gent": "девяти"},
                         "f": {"nomn": "девяти", "gent": "девяти"},
                         "n": {"nomn": "девяти", "gent": "девяти"}},
                     10: {"m": {"nomn": "десяти", "gent": "десяти"},
                          "f": {"nomn": "десяти", "gent": "десяти"},
                          "n": {"nomn": "десяти", "gent": "десяти"}},
                     11: {"m": {"nomn": "одиннадцати", "gent": "одиннадцати"},
                          "f": {"nomn": "одиннадцати", "gent": "одиннадцати"},
                          "n": {"nomn": "одиннадцати", "gent": "одиннадцати"}},
                     12: {"m": {"nomn": "двенадцати", "gent": "двенадцати"},
                          "f": {"nomn": "двенадцати", "gent": "двенадцати"},
                          "n": {"nomn": "двенадцати", "gent": "двенадцати"}},
                     13: {"m": {"nomn": "тринадцати", "gent": "тринадцати"},
                          "f": {"nomn": "тринадцати", "gent": "тринадцати"},
                          "n": {"nomn": "тринадцати", "gent": "тринадцати"}},
                     14: {"m": {"nomn": "четырнадцати", "gent": "четырнадцати"},
                          "f": {"nomn": "четырнадцати", "gent": "четырнадцати"},
                          "n": {"nomn": "четырнадцати", "gent": "четырнадцати"}},
                     15: {"m": {"nomn": "пятнадцати", "gent": "пятнадцати"},
                          "f": {"nomn": "пятнадцати", "gent": "пятнадцати"},
                          "n": {"nomn": "пятнадцати", "gent": "пятнадцати"}},
                     16: {"m": {"nomn": "шестнадцати", "gent": "шестнадцати"},
                          "f": {"nomn": "шестнадцати", "gent": "шестнадцати"},
                          "n": {"nomn": "шестнадцати", "gent": "шестнадцати"}},
                     17: {"m": {"nomn": "семнадцати", "gent": "семнадцати"},
                          "f": {"nomn": "семнадцати", "gent": "семнадцати"},
                          "n": {"nomn": "семнадцати", "gent": "семнадцати"}},
                     18: {"m": {"nomn": "восемнадцати", "gent": "восемнадцати"},
                          "f": {"nomn": "восемнадцати", "gent": "восемнадцати"},
                          "n": {"nomn": "восемнадцати", "gent": "восемнадцати"}},
                     19: {"m": {"nomn": "девятнадцати", "gent": "девятнадцати"},
                          "f": {"nomn": "девятнадцати", "gent": "девятнадцати"},
                          "n": {"nomn": "девятнадцати", "gent": "девятнадцати"}},
                     20: {"m": {"nomn": "двадцати", "gent": "двадцати"},
                          "f": {"nomn": "двадцати", "gent": "двадцати"},
                          "n": {"nomn": "двадцати", "gent": "двадцати"}},
                     30: {"m": {"nomn": "тридцати", "gent": "тридцати"},
                          "f": {"nomn": "тридцати", "gent": "тридцати"},
                          "n": {"nomn": "тридцати", "gent": "тридцати"}},
                     40: {"m": {"nomn": "сороко", "gent": "сороко"},
                          "f": {"nomn": "сороко", "gent": "сороко"},
                          "n": {"nomn": "сороко", "gent": "сороко"}},
                     50: {"m": {"nomn": "пятидесяти", "gent": "пятидесяти"},
                          "f": {"nomn": "пятидесяти", "gent": "пятидесяти"},
                          "n": {"nomn": "пятидесяти", "gent": "пятидесяти"}},
                     60: {"m": {"nomn": "шестидесяти", "gent": "шестидесяти"},
                          "f": {"nomn": "шестидесяти", "gent": "шестидесяти"},
                          "n": {"nomn": "шестидесяти", "gent": "шестидесяти"}},
                     70: {"m": {"nomn": "семидесяти", "gent": "семидесяти"},
                          "f": {"nomn": "семидесяти", "gent": "семидесяти"},
                          "n": {"nomn": "семидесяти", "gent": "семидесяти"}},
                     80: {"m": {"nomn": "восьмидесяти", "gent": "восьмидесяти"},
                          "f": {"nomn": "восьмидесяти", "gent": "восьмидесяти"},
                          "n": {"nomn": "восьмидесяти", "gent": "восьмидесяти"}},
                     90: {"m": {"nomn": "девяносто", "gent": "девяносто"},
                          "f": {"nomn": "девяносто", "gent": "девяносто"},
                          "n": {"nomn": "девяносто", "gent": "девяносто"}},
                     100: {"m": {"nomn": "сто", "gent": "сто"},
                           "f": {"nomn": "сто", "gent": "сто"},
                           "n": {"nomn": "сто", "gent": "сто"}},
                     200: {"m": {"nomn": "двухсот", "gent": "двухсот"},
                           "f": {"nomn": "двухсот", "gent": "двухсот"},
                           "n": {"nomn": "двухсот", "gent": "двухсот"}},
                     300: {"m": {"nomn": "трехсот", "gent": "трехсот"},
                           "f": {"nomn": "трехсот", "gent": "трехсот"},
                           "n": {"nomn": "трехсот", "gent": "трехсот"}},
                     400: {"m": {"nomn": "четырёхсот", "gent": "четырёхсот"},
                           "f": {"nomn": "четырёхсот", "gent": "четырёхсот"},
                           "n": {"nomn": "четырёхсот", "gent": "четырёхсот"}},
                     500: {"m": {"nomn": "пятисот", "gent": "пятисот"},
                           "f": {"nomn": "пятисот", "gent": "пятисот"},
                           "n": {"nomn": "пятисот", "gent": "пятисот"}},
                     600: {"m": {"nomn": "шестисот", "gent": "шестисот"},
                           "f": {"nomn": "шестисот", "gent": "шестисот"},
                           "n": {"nomn": "шестисот", "gent": "шестисот"}},
                     700: {"m": {"nomn": "семисот", "gent": "семисот"},
                           "f": {"nomn": "семисот", "gent": "семисот"},
                           "n": {"nomn": "семисот", "gent": "семисот"}},
                     800: {"m": {"nomn": "восьмисот", "gent": "восьмисот"},
                           "f": {"nomn": "восьмисот", "gent": "восьмисот"},
                           "n": {"nomn": "восьмисот", "gent": "восьмисот"}},
                     900: {"m": {"nomn": "девятисот", "gent": "девятисот"},
                           "f": {"nomn": "девятисот", "gent": "девятисот"},
                           "n": {"nomn": "девятисот", "gent": "девятисот"}}}
    structured = {0: {"m": {"nomn": "нулевой", "gent": "нулевого"},
                      "f": {"nomn": "нулевая", "gent": "нулевой"},
                      "n": {"nomn": "нулевое", "gent": "нулевого"}},
                  1: {"m": {"nomn": "первый", "gent": "первого"},
                      "f": {"nomn": "первая", "gent": "первой"},
                      "n": {"nomn": "первое", "gent": "первого"}},
                  2: {"m": {"nomn": "второй", "gent": "второго"},
                      "f": {"nomn": "вторая", "gent": "второй"},
                      "n": {"nomn": "второе", "gent": "второго"}},
                  3: {"m": {"nomn": "третий", "gent": "третьего"},
                      "f": {"nomn": "третья", "gent": "третьей"},
                      "n": {"nomn": "третье", "gent": "третьего"}},
                  4: {"m": {"nomn": "четвёртый", "gent": "четвёртого"},
                      "f": {"nomn": "четвёртая", "gent": "четвёртой"},
                      "n": {"nomn": "четвёртое", "gent": "четвёртого"}},
                  5: {"m": {"nomn": "пятый", "gent": "пятого"},
                      "f": {"nomn": "пятая", "gent": "пятой"},
                      "n": {"nomn": "пятое", "gent": "пятого"}},
                  6: {"m": {"nomn": "шестой", "gent": "шестого"},
                      "f": {"nomn": "шестая", "gent": "шестой"},
                      "n": {"nomn": "шестое", "gent": "шестого"}},
                  7: {"m": {"nomn": "седьмой", "gent": "седьмого"},
                      "f": {"nomn": "седьмая", "gent": "седьмой"},
                      "n": {"nomn": "седьмое", "gent": "седьмого"}},
                  8: {"m": {"nomn": "восьмой", "gent": "восьмого"},
                      "f": {"nomn": "восьмая", "gent": "восьмой"},
                      "n": {"nomn": "восьмое", "gent": "восьмого"}},
                  9: {"m": {"nomn": "девятый", "gent": "девятого"},
                      "f": {"nomn": "девятая", "gent": "девятой"},
                      "n": {"nomn": "девятое", "gent": "девятого"}},
                  10: {"m": {"nomn": "десятый", "gent": "десятого"},
                       "f": {"nomn": "десятая", "gent": "десятой"},
                       "n": {"nomn": "десятое", "gent": "десятого"}},
                  11: {"m": {"nomn": "одиннадцатый", "gent": "одиннадцатого"},
                       "f": {"nomn": "одиннадцатая", "gent": "одиннадцатой"},
                       "n": {"nomn": "одиннадцатое", "gent": "одиннадцатого"}},
                  12: {"m": {"nomn": "двенадцатый", "gent": "двенадцатого"},
                       "f": {"nomn": "двенадцатая", "gent": "двенадцатой"},
                       "n": {"nomn": "двенадцатое", "gent": "двенадцатого"}},
                  13: {"m": {"nomn": "тринадцатый", "gent": "тринадцатого"},
                       "f": {"nomn": "тринадцатая", "gent": "тринадцатой"},
                       "n": {"nomn": "тринадцатое", "gent": "тринадцатого"}},
                  14: {"m": {"nomn": "четырнадцатый", "gent": "четырнадцатого"},
                       "f": {"nomn": "четырнадцатая", "gent": "четырнадцатой"},
                       "n": {"nomn": "четырнадцатое", "gent": "четырнадцатого"}},
                  15: {"m": {"nomn": "пятнадцатый", "gent": "пятнадцатого"},
                       "f": {"nomn": "пятнадцатая", "gent": "пятнадцатой"},
                       "n": {"nomn": "пятнадцатое", "gent": "пятнадцатого"}},
                  16: {"m": {"nomn": "шестнадцатый", "gent": "шестнадцатого"},
                       "f": {"nomn": "шестнадцатая", "gent": "шестнадцатой"},
                       "n": {"nomn": "шестнадцатое", "gent": "шестнадцатого"}},
                  17: {"m": {"nomn": "семнадцатый", "gent": "семнадцатого"},
                       "f": {"nomn": "семнадцатая", "gent": "семнадцатой"},
                       "n": {"nomn": "семнадцатое", "gent": "семнадцатого"}},
                  18: {"m": {"nomn": "восемнадцатый", "gent": "восемнадцатого"},
                       "f": {"nomn": "восемнадцатая", "gent": "восемнадцатой"},
                       "n": {"nomn": "восемнадцатое", "gent": "восемнадцатого"}},
                  19: {"m": {"nomn": "девятнадцатый", "gent": "девятнадцатого"},
                       "f": {"nomn": "девятнадцатая", "gent": "девятнадцатой"},
                       "n": {"nomn": "девятнадцатое", "gent": "девятнадцатого"}},
                  20: {"m": {"nomn": "двадцатый", "gent": "двадцатого"},
                       "f": {"nomn": "двадцатая", "gent": "двадцатой"},
                       "n": {"nomn": "двадцатое", "gent": "двадцатого"}},
                  30: {"m": {"nomn": "тридцатый", "gent": "тридцатого"},
                       "f": {"nomn": "тридцатая", "gent": "тридцатой"},
                       "n": {"nomn": "тридцатое", "gent": "тридцатого"}},
                  40: {"m": {"nomn": "сороковой", "gent": "сорокового"},
                       "f": {"nomn": "сороковая", "gent": "сороковой"},
                       "n": {"nomn": "сороковое", "gent": "сорокового"}},
                  50: {"m": {"nomn": "пятидесятый", "gent": "пятидесятого"},
                       "f": {"nomn": "пятидесятая", "gent": "пятидесятой"},
                       "n": {"nomn": "пятидесятое", "gent": "пятидесятого"}},
                  60: {"m": {"nomn": "шестидесятый", "gent": "шестидесятого"},
                       "f": {"nomn": "шестидесятая", "gent": "шестидесятой"},
                       "n": {"nomn": "шестидесятое", "gent": "шестидесятого"}},
                  70: {"m": {"nomn": "семидесятый", "gent": "семидесятого"},
                       "f": {"nomn": "семидесятая", "gent": "семидесятой"},
                       "n": {"nomn": "семидесятое", "gent": "семидесятого"}},
                  80: {"m": {"nomn": "восьмидесятый", "gent": "восьмидесятого"},
                       "f": {"nomn": "восьмидесятая", "gent": "восьмидесятой"},
                       "n": {"nomn": "восьмидесятое", "gent": "восьмидесятого"}},
                  90: {"m": {"nomn": "девяностый", "gent": "девяностого"},
                       "f": {"nomn": "девяностая", "gent": "девяностой"},
                       "n": {"nomn": "девяностое", "gent": "девяностого"}},
                  100: {"m": {"nomn": "сотый", "gent": "сотого"},
                        "f": {"nomn": "сотая", "gent": "сотой"},
                        "n": {"nomn": "сотое", "gent": "сотого"}},
                  200: {"m": {"nomn": "двухсотый", "gent": "двухсотого"},
                        "f": {"nomn": "двухсотая", "gent": "двухсотой"},
                        "n": {"nomn": "двухсотое", "gent": "двухсотого"}},
                  300: {"m": {"nomn": "трёхсотый", "gent": "трёхсотого"},
                        "f": {"nomn": "трёхсотая", "gent": "трёхсотой"},
                        "n": {"nomn": "трёхсотое", "gent": "трёхсотого"}},
                  400: {"m": {"nomn": "четырёхсотый", "gent": "четырёхсотого"},
                        "f": {"nomn": "четырёхсотая", "gent": "четырёхсотой"},
                        "n": {"nomn": "четырёхсотое", "gent": "четырёхсотого"}},
                  500: {"m": {"nomn": "пятисотый", "gent": "пятисотого"},
                        "f": {"nomn": "пятисотая", "gent": "пятисотой"},
                        "n": {"nomn": "пятисотое", "gent": "пятисотого"}},
                  600: {"m": {"nomn": "шестисотый", "gent": "шестисотого"},
                        "f": {"nomn": "шестисотая", "gent": "шестисотой"},
                        "n": {"nomn": "шестисотое", "gent": "шестисотого"}},
                  700: {"m": {"nomn": "семисотый", "gent": "семисотого"},
                        "f": {"nomn": "семисотая", "gent": "семисотой"},
                        "n": {"nomn": "семисотое", "gent": "семисотого"}},
                  800: {"m": {"nomn": "восьмисотый", "gent": "восьмисотого"},
                        "f": {"nomn": "восьмисотая", "gent": "восьмисотой"},
                        "n": {"nomn": "восьмисотое", "gent": "восьмисотого"}},
                  900: {"m": {"nomn": "девятисотый", "gent": "девятисотого"},
                        "f": {"nomn": "девятисотая", "gent": "девятисотой"},
                        "n": {"nomn": "девятисотое", "gent": "девятисотого"}},
                  1000: {"m": {"nomn": "тысячный", "gent": "тысячного"},
                         "f": {"nomn": "тысячная", "gent": "тысячной"},
                         "n": {"nomn": "тысячное", "gent": "тысячного"}}
                  }
    end2form = {
        'ый': {'sex': 'm', 'case': 'nomn'},
        'ий': {'sex': 'm', 'case': 'nomn'},
        'й': {'sex': 'm', 'case': 'nomn'},
        'ое': {'sex': 'n', 'case': 'nomn'},
        'ая': {'sex': 'f', 'case': 'nomn'},
        'ого': {'sex': 'm', 'case': 'gent'},
        'го': {'sex': 'm', 'case': 'gent'},
        'ой': {'sex': 'f', 'case': 'gent'}
    }

    def __init__(self, num2text=None):
        self.num2text = num2text or Num2Text()
        self.re_any_ord = re.compile("[+-]?\d+")
        self.re_explicit_ord = re.compile("(\d+)\-?((?:о?го)|(?:ы?й)|(?:и?й)|(?:ой)|(?:ая)|(?:ое))")

    def _gen_sub1000(self, sub1000, data, sex, case):
        if sub1000 in data:
            result = data[sub1000][sex][case]
        else:
            sub100 = sub1000 % 100
            up100 = sub1000 - sub100
            if sub100 in data:
                result = data[sub100][sex][case]
                if up100:
                    result = self.num2text(up100) + " " + result
            else:
                sub10 = sub100 % 10
                up10 = sub100 - sub10
                result = self.num2text(up100 + up10) + " " + data[sub10][sex][case]
        return result

    def _gen_sub1000_orders(self, sub1000, data, sex, case):
        if sub1000 in data:
            result = data[sub1000][sex][case]
        else:
            sub100 = sub1000 % 100
            up100 = sub1000 - sub100
            sub10 = sub100 % 10
            up10 = sub100 - sub10
            if sub100 in data:
                result = data[sub100][sex][case]
                if up100:
                    result = data[up100][sex][case] + result
            else:
                result = ""
                for num in (up100, up10, sub10):
                    if num:
                        result += data[num][sex][case]
        return result.strip()

    def replace_everything_in_text(self, text, sex='m', case='nomn', only_explicit=False):
        repl = lambda m: " {} ".format(self(int(m.groups()[0]), **self.end2form[m.groups()[1]]))
        text = self.re_explicit_ord.sub(repl=repl, string=" {} ".format(text)).strip().replace("  ", " ").strip()
        if not only_explicit:
            repl = lambda x: " {} ".format(self(int(x.group()), sex=sex, case=case))
            text = self.re_any_ord.sub(repl=repl, string=" {} ".format(text)).strip().replace("  ", " ").strip()
        return text

    def __call__(self, num: int, sex="m", case="nomn"):
        is_negative = num < 0
        num = abs(num)
        if num in self.structured:
            result = self.structured[num][sex][case]
        else:
            sub1000 = num % 1000
            up1000 = num - sub1000
            if sub1000 == 0:
                order_lvl = (len(str(up1000)) - len(str(up1000).rstrip("0"))) // 3 * 3
                order = 10 ** order_lvl
                up_order = up1000 // order
                raw_for_one_word = up_order % 1000
                result = self._gen_sub1000_orders(raw_for_one_word, self.before_orders, sex, case) + \
                         self.orders[order_lvl][sex][case]
                if result.startswith("одно"):
                    result = result[4:]
                raw_for_num = (up_order - raw_for_one_word) * order
                if raw_for_num:
                    result = self.num2text(raw_for_num) + " " + result
                    if result.startswith(("один ", "одна ", "одно ")):
                        result = result[5:]
            else:
                result = self._gen_sub1000(sub1000, self.structured, sex, case)
                if up1000:
                    result = self.num2text(up1000) + " " + result
                if result.startswith(("один ", "одна ", "одно ")):
                    result = result[5:]
        if is_negative:
            result = "минус " + result
        return result


if __name__ == "__main__":
    num2ord = Num2Ordinal()
    t = "На улице 2 день хорошая погода"
    print("Проверим режим замены вхождений по всему тексту: \"{}\"".format(t))
    print(num2ord.replace_everything_in_text(t))
    print()

    print(num2ord(1, sex="m", case="nomn") + " человек на Луне")
    print(num2ord(87, sex="f", case="nomn") + " ракетка мира")
    print(num2ord(-3, sex="m", case="nomn") + " этаж")
    print(num2ord(8, sex="n", case="gent") + " сентября")
    print(num2ord(934674, sex="n", case="nomn") + " бредовое число")
    print(num2ord(1234000))
    print()
    print("вводите числа:")

    while True:
        t = int(input("> "))
        print(num2ord(t))
